import datetime
import logging
import random
import time

import pytest
import salt.utils.platform
import salt.utils.schedule
from tests.support.mock import MagicMock, patch

try:
    import dateutil.parser

    HAS_DATEUTIL_PARSER = True
except ImportError:
    HAS_DATEUTIL_PARSER = False


try:
    import croniter  # pylint: disable=unused-import

    HAS_CRONITER = True
except ImportError:
    HAS_CRONITER = False

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.skipif(
        HAS_DATEUTIL_PARSER is False,
        reason="The 'dateutil.parser' library is not available",
    ),
    pytest.mark.windows_whitelisted,
]


def _check_last_run(schedule, job_name, runtime=None):
    """
    Check that last_run time exists and
    is not the previous run time if prev_runtime
    is passed.
    """

    # The number of times to
    # check before bailing out.
    checks = 5

    status = schedule.job_status(job_name)
    count = 0
    while "_last_run" not in status:
        status = schedule.job_status(job_name)
        if count == checks:
            return False
        time.sleep(2)
        count += 1

    if runtime:
        count = 0
        status = schedule.job_status(job_name)
        while status["_last_run"] != runtime:
            if count == checks:
                return False
            status = schedule.job_status(job_name)
            time.sleep(2)
            count += 1
    return True


@pytest.mark.slow_test
def test_eval(schedule):
    """
    verify that scheduled job runs
    """

    job_name = "test_eval"
    job = {
        "schedule": {job_name: {"function": "test.ping", "when": "11/29/2017 4:00pm"}}
    }
    run_time2 = dateutil.parser.parse("11/29/2017 4:00pm")
    run_time1 = run_time2 - datetime.timedelta(seconds=1)

    # Add the job to the scheduler
    schedule.opts.update(job)

    # Evaluate 1 second before the run time
    schedule.eval(now=run_time1)
    ret = schedule.job_status(job_name)
    assert "_last_run" not in ret

    # Evaluate 1 second at the run time
    schedule.eval(now=run_time2)
    ret = schedule.job_status(job_name)
    assert ret["_last_run"] == run_time2


@pytest.mark.slow_test
def test_eval_multiple_whens(schedule):
    """
    verify that scheduled job runs
    """

    job_name = "test_eval_multiple_whens"
    job = {
        "schedule": {
            job_name: {
                "function": "test.ping",
                "when": ["11/29/2017 4:00pm", "11/29/2017 5:00pm"],
            }
        }
    }
    if salt.utils.platform.is_darwin():
        job["schedule"][job_name]["dry_run"] = True

    run_time1 = dateutil.parser.parse("11/29/2017 4:00pm")
    run_time2 = dateutil.parser.parse("11/29/2017 5:00pm")

    # Add the job to the scheduler
    schedule.opts.update(job)

    # Evaluate run time1
    schedule.eval(now=run_time1)

    # Give the job a chance to finish
    assert _check_last_run(schedule, job_name, run_time1)

    # Evaluate run time2
    schedule.eval(now=run_time2)

    # Give the job a chance to finish and check
    assert _check_last_run(schedule, job_name, run_time2)


@pytest.mark.slow_test
def test_eval_whens(schedule):
    """
    verify that scheduled job runs
    """

    schedule.opts["grains"]["whens"] = {"tea time": "11/29/2017 12:00pm"}

    job_name = "test_eval_whens"
    job = {"schedule": {job_name: {"function": "test.ping", "when": "tea time"}}}
    run_time = dateutil.parser.parse("11/29/2017 12:00pm")

    # Add the job to the scheduler
    schedule.opts.update(job)

    # Evaluate run time1
    schedule.eval(now=run_time)
    assert _check_last_run(schedule, job_name, run_time)


@pytest.mark.slow_test
def test_eval_loop_interval(schedule):
    """
    verify that scheduled job runs
    """

    job_name = "test_eval_loop_interval"
    job = {
        "schedule": {job_name: {"function": "test.ping", "when": "11/29/2017 4:00pm"}}
    }
    # 30 second loop interval
    LOOP_INTERVAL = random.randint(30, 59)
    schedule.opts["loop_interval"] = LOOP_INTERVAL

    run_time2 = dateutil.parser.parse("11/29/2017 4:00pm")

    # Add the job to the scheduler
    schedule.opts.update(job)

    # Evaluate 1 second at the run time
    schedule.eval(now=run_time2 + datetime.timedelta(seconds=LOOP_INTERVAL))

    assert _check_last_run(
        schedule, job_name, run_time2 + datetime.timedelta(seconds=LOOP_INTERVAL)
    )


@pytest.mark.slow_test
def test_eval_multiple_whens_loop_interval(schedule):
    """
    verify that scheduled job runs
    """

    job_name = "test_eval_multiple_whens_loop_interval"
    job = {
        "schedule": {
            job_name: {
                "function": "test.ping",
                "when": ["11/29/2017 4:00pm", "11/29/2017 5:00pm"],
            }
        }
    }
    if salt.utils.platform.is_darwin():
        job["schedule"][job_name]["dry_run"] = True

    # 30 second loop interval
    LOOP_INTERVAL = random.randint(30, 59)
    schedule.opts["loop_interval"] = LOOP_INTERVAL

    run_time1 = dateutil.parser.parse("11/29/2017 4:00pm") + datetime.timedelta(
        seconds=LOOP_INTERVAL
    )
    run_time2 = dateutil.parser.parse("11/29/2017 5:00pm") + datetime.timedelta(
        seconds=LOOP_INTERVAL
    )

    # Add the job to the scheduler
    schedule.opts.update(job)

    # Evaluate 1 second at the run time
    schedule.eval(now=run_time1)

    # Give the job a chance to finish
    assert _check_last_run(schedule, job_name, run_time1)

    # Evaluate 1 second at the run time
    schedule.eval(now=run_time2)

    # Give the job a chance to finish
    assert _check_last_run(schedule, job_name, run_time2)


@pytest.mark.slow_test
def test_eval_once(schedule):
    """
    verify that scheduled job runs
    """

    job_name = "test_once"
    job = {
        "schedule": {job_name: {"function": "test.ping", "once": "2017-12-13T13:00:00"}}
    }
    run_time = dateutil.parser.parse("12/13/2017 1:00pm")

    # Add the job to the scheduler
    schedule.opts["schedule"] = {}
    schedule.opts.update(job)

    # Evaluate 1 second at the run time
    schedule.eval(now=run_time)
    assert _check_last_run(schedule, job_name, run_time)


@pytest.mark.slow_test
def test_eval_once_loop_interval(schedule):
    """
    verify that scheduled job runs
    """

    job_name = "test_eval_once_loop_interval"
    job = {
        "schedule": {job_name: {"function": "test.ping", "once": "2017-12-13T13:00:00"}}
    }
    # Randomn second loop interval
    LOOP_INTERVAL = random.randint(0, 59)
    schedule.opts["loop_interval"] = LOOP_INTERVAL

    # Run the job at the right plus LOOP_INTERVAL
    run_time = dateutil.parser.parse("12/13/2017 1:00pm") + datetime.timedelta(
        seconds=LOOP_INTERVAL
    )

    # Add the job to the scheduler
    schedule.opts.update(job)

    # Evaluate at the run time
    schedule.eval(now=run_time)
    assert _check_last_run(schedule, job_name, run_time)


@pytest.mark.skipif(not HAS_CRONITER, reason="Cannot find croniter python module")
def test_eval_cron(schedule):
    """
    verify that scheduled job runs
    """

    job_name = "test_eval_cron"
    job = {"schedule": {job_name: {"function": "test.ping", "cron": "0 16 29 11 *"}}}

    # Add the job to the scheduler
    schedule.opts.update(job)

    run_time = dateutil.parser.parse("11/29/2017 4:00pm")
    with patch("croniter.croniter.get_next", MagicMock(return_value=run_time)):
        schedule.eval(now=run_time)
    assert _check_last_run(schedule, job_name, run_time)


@pytest.mark.skipif(not HAS_CRONITER, reason="Cannot find croniter python module")
def test_eval_cron_loop_interval(schedule):
    """
    verify that scheduled job runs
    """

    job_name = "test_eval_cron_loop_interval"
    job = {"schedule": {job_name: {"function": "test.ping", "cron": "0 16 29 11 *"}}}
    # Randomn second loop interval
    LOOP_INTERVAL = random.randint(0, 59)
    schedule.opts["loop_interval"] = LOOP_INTERVAL

    # Add the job to the scheduler
    schedule.opts.update(job)

    run_time = dateutil.parser.parse("11/29/2017 4:00pm")
    with patch("croniter.croniter.get_next", MagicMock(return_value=run_time)):
        schedule.eval(now=run_time)
    assert _check_last_run(schedule, job_name, run_time)


@pytest.mark.slow_test
def test_eval_until(schedule):
    """
    verify that scheduled job is skipped once the current
    time reaches the specified until time
    """

    job_name = "test_eval_until"
    job = {
        "schedule": {
            job_name: {
                "function": "test.ping",
                "hours": "1",
                "until": "11/29/2017 5:00pm",
            }
        }
    }

    if salt.utils.platform.is_darwin():
        job["schedule"][job_name]["dry_run"] = True

    # Add job to schedule
    schedule.delete_job("test_eval_until")
    schedule.opts.update(job)

    # eval at 2:00pm to prime, simulate minion start up.
    run_time = dateutil.parser.parse("11/29/2017 2:00pm")
    schedule.eval(now=run_time)

    # eval at 3:00pm, will run.
    run_time = dateutil.parser.parse("11/29/2017 3:00pm")
    schedule.eval(now=run_time)

    # Give the job a chance to finish
    assert _check_last_run(schedule, job_name, run_time)

    # eval at 4:00pm, will run.
    run_time = dateutil.parser.parse("11/29/2017 4:00pm")
    schedule.eval(now=run_time)

    # Give the job a chance to finish
    assert _check_last_run(schedule, job_name, run_time)

    # eval at 5:00pm, will not run
    run_time = dateutil.parser.parse("11/29/2017 5:00pm")
    schedule.eval(now=run_time)

    ret = schedule.job_status(job_name)

    assert ret["_skip_reason"] == "until_passed"
    assert ret["_skipped_time"] == run_time


@pytest.mark.slow_test
def test_eval_after(schedule):
    """
    verify that scheduled job is skipped until after the specified
    time has been reached.
    """

    job_name = "test_eval_after"
    job = {
        "schedule": {
            job_name: {
                "function": "test.ping",
                "hours": "1",
                "after": "11/29/2017 5:00pm",
            }
        }
    }

    # Add job to schedule
    schedule.opts.update(job)

    # eval at 2:00pm to prime, simulate minion start up.
    run_time = dateutil.parser.parse("11/29/2017 2:00pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)

    # eval at 3:00pm, will not run.
    run_time = dateutil.parser.parse("11/29/2017 3:00pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert ret["_skip_reason"] == "after_not_passed"
    assert ret["_skipped_time"] == run_time

    # eval at 4:00pm, will not run.
    run_time = dateutil.parser.parse("11/29/2017 4:00pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert ret["_skip_reason"] == "after_not_passed"
    assert ret["_skipped_time"] == run_time

    # eval at 5:00pm, will not run
    run_time = dateutil.parser.parse("11/29/2017 5:00pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert ret["_skip_reason"] == "after_not_passed"
    assert ret["_skipped_time"] == run_time

    # eval at 6:00pm, will run
    run_time = dateutil.parser.parse("11/29/2017 6:00pm")
    schedule.eval(now=run_time)
    assert _check_last_run(schedule, job_name, run_time)


@pytest.mark.slow_test
def test_eval_enabled(schedule):
    """
    verify that scheduled job does not run
    """

    job_name = "test_eval_enabled"
    job = {
        "schedule": {
            "enabled": True,
            job_name: {"function": "test.ping", "when": "11/29/2017 4:00pm"},
        }
    }
    run_time1 = dateutil.parser.parse("11/29/2017 4:00pm")

    # Add the job to the scheduler
    schedule.opts.update(job)

    # Evaluate 1 second at the run time
    schedule.eval(now=run_time1)
    assert _check_last_run(schedule, job_name, run_time1)


@pytest.mark.slow_test
def test_eval_enabled_key(schedule):
    """
    verify that scheduled job runs
    when the enabled key is in place
    https://github.com/saltstack/salt/issues/47695
    """

    job_name = "test_eval_enabled_key"
    job = {
        "schedule": {
            "enabled": True,
            job_name: {"function": "test.ping", "when": "11/29/2017 4:00pm"},
        }
    }
    run_time2 = dateutil.parser.parse("11/29/2017 4:00pm")
    run_time1 = run_time2 - datetime.timedelta(seconds=1)

    # Add the job to the scheduler
    schedule.opts.update(job)

    # Evaluate 1 second before the run time
    schedule.eval(now=run_time1)
    ret = schedule.job_status("test_eval_enabled_key")
    assert "_last_run" not in ret

    # Evaluate 1 second at the run time
    schedule.eval(now=run_time2)
    assert _check_last_run(schedule, job_name, run_time2)


def test_eval_disabled(schedule):
    """
    verify that scheduled job does not run
    """

    job_name = "test_eval_disabled"
    job = {
        "schedule": {
            "enabled": False,
            job_name: {"function": "test.ping", "when": "11/29/2017 4:00pm"},
        }
    }
    run_time1 = dateutil.parser.parse("11/29/2017 4:00pm")

    # Add the job to the scheduler
    schedule.opts.update(job)

    # Evaluate 1 second at the run time
    schedule.eval(now=run_time1)
    ret = schedule.job_status(job_name)
    assert "_last_run" not in ret
    assert ret["_skip_reason"] == "disabled"

    # Ensure job data still matches
    assert ret == job["schedule"][job_name]


def test_eval_global_disabled_job_enabled(schedule):
    """
    verify that scheduled job does not run
    """

    job_name = "test_eval_global_disabled"
    job = {
        "schedule": {
            "enabled": False,
            job_name: {
                "function": "test.ping",
                "when": "11/29/2017 4:00pm",
                "enabled": True,
            },
        }
    }
    run_time1 = dateutil.parser.parse("11/29/2017 4:00pm")

    # Add the job to the scheduler
    schedule.opts.update(job)

    # Evaluate 1 second at the run time
    schedule.eval(now=run_time1)
    ret = schedule.job_status(job_name)
    assert "_last_run" not in ret
    assert ret["_skip_reason"] == "disabled"

    # Ensure job is still enabled
    assert ret["enabled"]


@pytest.mark.slow_test
def test_eval_run_on_start(schedule):
    """
    verify that scheduled job is run when minion starts
    """

    job_name = "test_eval_run_on_start"
    job = {
        "schedule": {
            job_name: {"function": "test.ping", "hours": "1", "run_on_start": True}
        }
    }

    # Add job to schedule
    schedule.opts.update(job)

    # eval at 2:00pm, will run.
    run_time = dateutil.parser.parse("11/29/2017 2:00pm")
    schedule.eval(now=run_time)
    assert _check_last_run(schedule, job_name, run_time)

    # eval at 3:00pm, will run.
    run_time = dateutil.parser.parse("11/29/2017 3:00pm")
    schedule.eval(now=run_time)
    assert _check_last_run(schedule, job_name, run_time)


@pytest.mark.slow_test
def test_eval_splay(schedule):
    """
    verify that scheduled job runs with splayed time
    """

    job_name = "job_eval_splay"
    job = {
        "schedule": {
            job_name: {"function": "test.ping", "seconds": "30", "splay": "10"}
        }
    }

    # Add job to schedule
    schedule.opts.update(job)

    with patch("random.randint", MagicMock(return_value=10)):
        # eval at 2:00pm to prime, simulate minion start up.
        run_time = dateutil.parser.parse("11/29/2017 2:00pm")
        schedule.eval(now=run_time)

        # eval at 2:00:40pm, will run.
        run_time = dateutil.parser.parse("11/29/2017 2:00:40pm")
        schedule.eval(now=run_time)
        assert _check_last_run(schedule, job_name, run_time)


@pytest.mark.slow_test
def test_eval_splay_range(schedule):
    """
    verify that scheduled job runs with splayed time
    """

    job_name = "job_eval_splay_range"
    job = {
        "schedule": {
            job_name: {
                "function": "test.ping",
                "seconds": "30",
                "splay": {"start": 5, "end": 10},
            }
        }
    }

    # Add job to schedule
    schedule.opts.update(job)

    with patch("random.randint", MagicMock(return_value=10)):
        # eval at 2:00pm to prime, simulate minion start up.
        run_time = dateutil.parser.parse("11/29/2017 2:00pm")
        schedule.eval(now=run_time)

        # eval at 2:00:40pm, will run.
        run_time = dateutil.parser.parse("11/29/2017 2:00:40pm")
        schedule.eval(now=run_time)
        assert _check_last_run(schedule, job_name, run_time)


@pytest.mark.slow_test
def test_eval_splay_global(schedule):
    """
    verify that scheduled job runs with splayed time
    """

    job_name = "job_eval_splay_global"
    job = {
        "schedule": {
            "splay": {"start": 5, "end": 10},
            job_name: {"function": "test.ping", "seconds": "30"},
        }
    }

    # Add job to schedule
    schedule.opts.update(job)

    with patch("random.randint", MagicMock(return_value=10)):
        # eval at 2:00pm to prime, simulate minion start up.
        run_time = dateutil.parser.parse("11/29/2017 2:00pm")
        schedule.eval(now=run_time)

        # eval at 2:00:40pm, will run.
        run_time = dateutil.parser.parse("11/29/2017 2:00:40pm")
        schedule.eval(now=run_time)
        assert _check_last_run(schedule, job_name, run_time)


@pytest.mark.slow_test
def test_eval_seconds(schedule):
    """
    verify that scheduled job run mutiple times with seconds
    """

    with patch.dict(schedule.opts, {"run_schedule_jobs_in_background": False}):
        job_name = "job_eval_seconds"
        job = {"schedule": {job_name: {"function": "test.ping", "seconds": "30"}}}

        if salt.utils.platform.is_darwin():
            job["schedule"][job_name]["dry_run"] = True

        # Add job to schedule
        schedule.opts.update(job)

        # eval at 2:00pm to prime, simulate minion start up.
        run_time = dateutil.parser.parse("11/29/2017 2:00pm")
        next_run_time = run_time + datetime.timedelta(seconds=30)
        jids = schedule.eval(now=run_time)
        assert len(jids) == 0
        ret = schedule.job_status(job_name)
        assert ret["_next_fire_time"] == next_run_time

        # eval at 2:00:01pm, will not run.
        run_time = dateutil.parser.parse("11/29/2017 2:00:01pm")
        jids = schedule.eval(now=run_time)
        assert len(jids) == 0
        ret = schedule.job_status(job_name)
        assert "_last_run" not in ret
        assert ret["_next_fire_time"] == next_run_time

        # eval at 2:00:30pm, will run.
        run_time = dateutil.parser.parse("11/29/2017 2:00:30pm")
        next_run_time = run_time + datetime.timedelta(seconds=30)
        jids = schedule.eval(now=run_time)
        assert len(jids) == 1
        ret = schedule.job_status(job_name)
        assert ret["_last_run"] == run_time
        assert ret["_next_fire_time"] == next_run_time

        # eval at 2:01:00pm, will run.
        run_time = dateutil.parser.parse("11/29/2017 2:01:00pm")
        next_run_time = run_time + datetime.timedelta(seconds=30)
        jids = schedule.eval(now=run_time)
        assert len(jids) == 1
        ret = schedule.job_status(job_name)
        assert ret["_last_run"] == run_time
        assert ret["_next_fire_time"] == next_run_time

        # eval at 2:01:30pm, will run.
        run_time = dateutil.parser.parse("11/29/2017 2:01:30pm")
        next_run_time = run_time + datetime.timedelta(seconds=30)
        jids = schedule.eval(now=run_time)
        assert len(jids) == 1
        ret = schedule.job_status(job_name)
        assert ret["_last_run"] == run_time
        assert ret["_next_fire_time"] == next_run_time


@pytest.mark.slow_test
def test_eval_minutes(schedule):
    """
    verify that scheduled job run mutiple times with minutes
    """

    with patch.dict(schedule.opts, {"run_schedule_jobs_in_background": False}):
        job_name = "job_eval_minutes"
        job = {"schedule": {job_name: {"function": "test.ping", "minutes": "30"}}}

        if salt.utils.platform.is_darwin():
            job["schedule"][job_name]["dry_run"] = True

        # Add job to schedule
        schedule.opts.update(job)

        # eval at 2:00pm to prime, simulate minion start up.
        run_time = dateutil.parser.parse("11/29/2017 2:00pm")
        next_run_time = run_time + datetime.timedelta(minutes=30)
        jids = schedule.eval(now=run_time)
        assert len(jids) == 0
        ret = schedule.job_status(job_name)
        assert ret["_next_fire_time"] == next_run_time

        # eval at 2:00:01pm, will not run.
        run_time = dateutil.parser.parse("11/29/2017 2:00:01pm")
        jids = schedule.eval(now=run_time)
        assert len(jids) == 0
        ret = schedule.job_status(job_name)
        assert "_last_run" not in ret
        assert ret["_next_fire_time"] == next_run_time

        # eval at 2:30:00pm, will run.
        run_time = dateutil.parser.parse("11/29/2017 2:30:00pm")
        jids = schedule.eval(now=run_time)
        assert len(jids) == 1
        ret = schedule.job_status(job_name)
        assert ret["_last_run"] == run_time

        # eval at 3:00:00pm, will run.
        run_time = dateutil.parser.parse("11/29/2017 3:00:00pm")
        jids = schedule.eval(now=run_time)
        assert len(jids) == 1
        ret = schedule.job_status(job_name)
        assert ret["_last_run"] == run_time

        # eval at 3:30:00pm, will run.
        run_time = dateutil.parser.parse("11/29/2017 3:30:00pm")
        jids = schedule.eval(now=run_time)
        assert len(jids) == 1
        ret = schedule.job_status(job_name)
        assert ret["_last_run"] == run_time


@pytest.mark.slow_test
def test_eval_hours(schedule):
    """
    verify that scheduled job run mutiple times with hours
    """

    with patch.dict(schedule.opts, {"run_schedule_jobs_in_background": False}):
        job_name = "job_eval_hours"
        job = {"schedule": {job_name: {"function": "test.ping", "hours": "2"}}}

        if salt.utils.platform.is_darwin():
            job["schedule"][job_name]["dry_run"] = True

        # Add job to schedule
        schedule.opts.update(job)

        # eval at 2:00pm to prime, simulate minion start up.
        run_time = dateutil.parser.parse("11/29/2017 2:00pm")
        next_run_time = run_time + datetime.timedelta(hours=2)
        jids = schedule.eval(now=run_time)
        assert len(jids) == 0
        ret = schedule.job_status(job_name)
        assert ret["_next_fire_time"] == next_run_time

        # eval at 2:00:01pm, will not run.
        run_time = dateutil.parser.parse("11/29/2017 2:00:01pm")
        jids = schedule.eval(now=run_time)
        assert len(jids) == 0
        ret = schedule.job_status(job_name)
        assert "_last_run" not in ret
        assert ret["_next_fire_time"] == next_run_time

        # eval at 4:00:00pm, will run.
        run_time = dateutil.parser.parse("11/29/2017 4:00:00pm")
        jids = schedule.eval(now=run_time)
        assert len(jids) == 1
        ret = schedule.job_status(job_name)
        assert ret["_last_run"] == run_time

        # eval at 6:00:00pm, will run.
        run_time = dateutil.parser.parse("11/29/2017 6:00:00pm")
        jids = schedule.eval(now=run_time)
        assert len(jids) == 1

        ret = schedule.job_status(job_name)
        assert ret["_last_run"] == run_time

        # eval at 8:00:00pm, will run.
        run_time = dateutil.parser.parse("11/29/2017 8:00:00pm")
        pids = schedule.eval(now=run_time)
        assert len(jids) == 1

        ret = schedule.job_status(job_name)
        assert ret["_last_run"] == run_time


@pytest.mark.slow_test
def test_eval_days(schedule):
    """
    verify that scheduled job run mutiple times with days
    """

    job_name = "job_eval_days"
    job = {
        "schedule": {job_name: {"function": "test.ping", "days": "2", "dry_run": True}}
    }

    if salt.utils.platform.is_darwin():
        job["schedule"][job_name]["dry_run"] = True

    # Add job to schedule
    schedule.opts.update(job)

    # eval at 11/23/2017 2:00pm to prime, simulate minion start up.
    run_time = dateutil.parser.parse("11/23/2017 2:00pm")
    next_run_time = run_time + datetime.timedelta(days=2)
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert ret["_next_fire_time"] == next_run_time

    # eval at 11/25/2017 2:00:00pm, will run.
    run_time = dateutil.parser.parse("11/25/2017 2:00:00pm")
    next_run_time = run_time + datetime.timedelta(days=2)
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)
    assert ret["_last_run"] == run_time
    assert ret["_next_fire_time"] == next_run_time

    # eval at 11/26/2017 2:00:00pm, will not run.
    run_time = dateutil.parser.parse("11/26/2017 2:00:00pm")
    last_run_time = run_time - datetime.timedelta(days=1)
    schedule.eval(now=run_time)

    # Give the job a chance to finish
    time.sleep(5)

    ret = schedule.job_status(job_name)

    assert ret["_last_run"] == last_run_time
    assert ret["_next_fire_time"] == next_run_time

    # eval at 11/27/2017 2:00:00pm, will run.
    run_time = dateutil.parser.parse("11/27/2017 2:00:00pm")
    next_run_time = run_time + datetime.timedelta(days=2)
    schedule.eval(now=run_time)

    # Give the job a chance to finish
    time.sleep(5)

    ret = schedule.job_status(job_name)

    assert ret["_last_run"] == run_time
    assert ret["_next_fire_time"] == next_run_time

    # eval at 11/28/2017 2:00:00pm, will not run.
    run_time = dateutil.parser.parse("11/28/2017 2:00:00pm")
    last_run_time = run_time - datetime.timedelta(days=1)
    schedule.eval(now=run_time)

    # Give the job a chance to finish
    time.sleep(5)

    ret = schedule.job_status(job_name)

    assert ret["_last_run"] == last_run_time
    assert ret["_next_fire_time"] == next_run_time

    # eval at 11/29/2017 2:00:00pm, will run.
    run_time = dateutil.parser.parse("11/29/2017 2:00:00pm")
    next_run_time = run_time + datetime.timedelta(days=2)
    schedule.eval(now=run_time)

    # Give the job a chance to finish
    time.sleep(5)

    ret = schedule.job_status(job_name)

    assert ret["_last_run"] == run_time
    assert ret["_next_fire_time"] == next_run_time


@pytest.mark.slow_test
def test_eval_when_splay(schedule):
    """
    verify that scheduled job runs
    """

    job_name = "test_eval_when_splay"
    splay = 300
    job = {
        "schedule": {
            job_name: {
                "function": "test.ping",
                "when": "11/29/2017 4:00pm",
                "splay": splay,
            }
        }
    }
    run_time1 = dateutil.parser.parse("11/29/2017 4:00pm")
    run_time2 = run_time1 + datetime.timedelta(seconds=splay)
    run_time3 = run_time2 + datetime.timedelta(seconds=1)

    # Add the job to the scheduler
    schedule.opts.update(job)

    with patch("random.randint", MagicMock(return_value=splay)):
        # Evaluate to prime
        run_time = dateutil.parser.parse("11/29/2017 3:00pm")
        schedule.eval(now=run_time)
        ret = schedule.job_status(job_name)

        # Evaluate at expected runtime1, should not run
        schedule.eval(now=run_time1)
        ret = schedule.job_status(job_name)
        assert "_last_run" not in ret

        # Evaluate at expected runtime2, should run
        schedule.eval(now=run_time2)
        ret = schedule.job_status(job_name)
        assert ret["_last_run"] == run_time2

        # Evaluate at expected runtime3, should not run
        # _next_fire_time should be None
        schedule.eval(now=run_time3)
        ret = schedule.job_status(job_name)
        assert ret["_last_run"] == run_time2
        assert ret["_next_fire_time"] is None


def test_eval_when_splay_in_past(schedule):
    """
    verify that scheduled job runs
    """

    job_name = "test_eval_when_splay_in_past"
    splay = 300
    job = {
        "schedule": {
            job_name: {
                "function": "test.ping",
                "when": ["11/29/2017 6:00am"],
                "splay": splay,
            }
        }
    }
    run_time1 = dateutil.parser.parse("11/29/2017 4:00pm")

    # Add the job to the scheduler
    schedule.opts.update(job)

    # Evaluate to prime
    run_time = dateutil.parser.parse("11/29/2017 3:00pm")
    schedule.eval(now=run_time)
    ret = schedule.job_status(job_name)

    # Evaluate at expected runtime1, should not run
    # and _next_fire_time should be None
    schedule.eval(now=run_time1)
    ret = schedule.job_status(job_name)
    assert "_last_run" not in ret
    assert ret["_next_fire_time"] is None



"""
unit tests for the cache runner
"""


import salt.runners.cache as cache
import salt.utils.master
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase


class CacheTest(TestCase, LoaderModuleMockMixin):
    """
    Validate the cache runner
    """

    def setup_loader_modules(self):
        return {
            cache: {
                "__opts__": {
                    "cache": "localfs",
                    "pki_dir": RUNTIME_VARS.TMP,
                    "key_cache": True,
                }
            }
        }

    def test_grains(self):
        """
        test cache.grains runner
        """
        mock_minion = ["Larry"]
        mock_ret = {}
        self.assertEqual(cache.grains(tgt="*", minion=mock_minion), mock_ret)

        mock_data = "grain stuff"

        class MockMaster:
            def __init__(self, *args, **kwargs):
                pass

            def get_minion_grains(self):
                return mock_data

        with patch.object(salt.utils.master, "MasterPillarUtil", MockMaster):
            self.assertEqual(cache.grains(tgt="*"), mock_data)



"""
Module to integrate with the returner system and retrieve data sent to a salt returner
"""

import salt.loader


def get_jid(returner, jid):
    """
    Return the information for a specified job id

    CLI Example:

    .. code-block:: bash

        salt '*' ret.get_jid redis 20421104181954700505
    """
    returners = salt.loader.returners(__opts__, __salt__)
    return returners["{}.get_jid".format(returner)](jid)


def get_fun(returner, fun):
    """
    Return info about last time fun was called on each minion

    CLI Example:

    .. code-block:: bash

        salt '*' ret.get_fun mysql network.interfaces
    """
    returners = salt.loader.returners(__opts__, __salt__)
    return returners["{}.get_fun".format(returner)](fun)


def get_jids(returner):
    """
    Return a list of all job ids

    CLI Example:

    .. code-block:: bash

        salt '*' ret.get_jids mysql
    """
    returners = salt.loader.returners(__opts__, __salt__)
    return returners["{}.get_jids".format(returner)]()


def get_minions(returner):
    """
    Return a list of all minions

    CLI Example:

    .. code-block:: bash

        salt '*' ret.get_minions mysql
    """
    returners = salt.loader.returners(__opts__, __salt__)
    return returners["{}.get_minions".format(returner)]()



"""
Feature flags
"""
import logging

log = logging.getLogger(__name__)


class Features:
    def __init__(self, _features=None):
        if _features is None:
            self.features = {}
        else:
            self.features = _features
        self.setup = False

    def setup_features(self, opts):
        if not self.setup:
            self.features.update(opts.get("features", {}))
        else:
            log.warn("Features already setup")

    def get(self, key, default=None):
        return self.features.get(key, default)


features = Features()
setup_features = features.setup_features



"""
Module for running vmadm command on SmartOS
"""

import logging
import os

import salt.utils.args
import salt.utils.files
import salt.utils.json
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
from salt.utils.odict import OrderedDict

try:
    from shlex import quote as _quote_args  # pylint: disable=E0611
except ImportError:
    from pipes import quote as _quote_args


log = logging.getLogger(__name__)

# Function aliases
__func_alias__ = {"list_vms": "list"}

# Define the module's virtual name
__virtualname__ = "vmadm"


def __virtual__():
    """
    Provides vmadm on SmartOS
    """
    if (
        salt.utils.platform.is_smartos_globalzone()
        and salt.utils.path.which("vmadm")
        and salt.utils.path.which("zfs")
    ):
        return __virtualname__
    return (
        False,
        "{} module can only be loaded on SmartOS compute nodes".format(__virtualname__),
    )


def _exit_status(retcode):
    """
    Translate exit status of vmadm
    """
    ret = {0: "Successful completion.", 1: "An error occurred.", 2: "Usage error."}[
        retcode
    ]
    return ret


def _create_update_from_file(mode="create", uuid=None, path=None):
    """
    Create vm from file
    """
    ret = {}
    if not os.path.isfile(path) or path is None:
        ret["Error"] = "File ({}) does not exists!".format(path)
        return ret
    # vmadm validate create|update [-f <filename>]
    cmd = "vmadm validate {mode} {brand} -f {path}".format(
        mode=mode, brand=get(uuid)["brand"] if uuid is not None else "", path=path
    )
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        if "stderr" in res:
            if res["stderr"][0] == "{":
                ret["Error"] = salt.utils.json.loads(res["stderr"])
            else:
                ret["Error"] = res["stderr"]
        return ret
    # vmadm create|update [-f <filename>]
    cmd = "vmadm {mode} {uuid} -f {path}".format(
        mode=mode, uuid=uuid if uuid is not None else "", path=path
    )
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        if "stderr" in res:
            if res["stderr"][0] == "{":
                ret["Error"] = salt.utils.json.loads(res["stderr"])
            else:
                ret["Error"] = res["stderr"]
        return ret
    else:
        if res["stderr"].startswith("Successfully created VM"):
            return res["stderr"][24:]
    return True


def _create_update_from_cfg(mode="create", uuid=None, vmcfg=None):
    """
    Create vm from configuration
    """
    ret = {}

    # write json file
    vmadm_json_file = __salt__["temp.file"](prefix="vmadm-")
    with salt.utils.files.fopen(vmadm_json_file, "w") as vmadm_json:
        salt.utils.json.dump(vmcfg, vmadm_json)

    # vmadm validate create|update [-f <filename>]
    cmd = "vmadm validate {mode} {brand} -f {vmadm_json_file}".format(
        mode=mode,
        brand=get(uuid)["brand"] if uuid is not None else "",
        vmadm_json_file=vmadm_json_file,
    )
    res = __salt__["cmd.run_all"](cmd, python_shell=True)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        if "stderr" in res:
            if res["stderr"][0] == "{":
                ret["Error"] = salt.utils.json.loads(res["stderr"])
            else:
                ret["Error"] = res["stderr"]
        return ret
    # vmadm create|update [-f <filename>]
    cmd = "vmadm {mode} {uuid} -f {vmadm_json_file}".format(
        mode=mode,
        uuid=uuid if uuid is not None else "",
        vmadm_json_file=vmadm_json_file,
    )
    res = __salt__["cmd.run_all"](cmd, python_shell=True)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        if "stderr" in res:
            if res["stderr"][0] == "{":
                ret["Error"] = salt.utils.json.loads(res["stderr"])
            else:
                ret["Error"] = res["stderr"]
        return ret
    else:
        # cleanup json file (only when successful to help troubleshooting)
        salt.utils.files.safe_rm(vmadm_json_file)

        # return uuid
        if res["stderr"].startswith("Successfully created VM"):
            return res["stderr"][24:]

    return True


def start(vm, options=None, key="uuid"):
    """
    Start a vm

    vm : string
        vm to be started
    options : string
        optional additional options
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.start 186da9ab-7392-4f55-91a5-b8f1fe770543
        salt '*' vmadm.start 186da9ab-7392-4f55-91a5-b8f1fe770543 'order=c,once=d cdrom=/path/to/image.iso,ide'
        salt '*' vmadm.start vm=nacl key=alias
        salt '*' vmadm.start vm=nina.example.org key=hostname
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup("{}={}".format(key, vm), one=True)
    if "Error" in vm:
        return vm
    # vmadm start <uuid> [option=value ...]
    cmd = "vmadm start {uuid} {options}".format(
        uuid=vm, options=options if options else ""
    )
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def stop(vm, force=False, key="uuid"):
    """
    Stop a vm

    vm : string
        vm to be stopped
    force : boolean
        force stop of vm if true
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.stop 186da9ab-7392-4f55-91a5-b8f1fe770543
        salt '*' vmadm.stop 186da9ab-7392-4f55-91a5-b8f1fe770543 True
        salt '*' vmadm.stop vm=nacl key=alias
        salt '*' vmadm.stop vm=nina.example.org key=hostname
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup("{}={}".format(key, vm), one=True)
    if "Error" in vm:
        return vm
    # vmadm stop <uuid> [-F]
    cmd = "vmadm stop {force} {uuid}".format(force="-F" if force else "", uuid=vm)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = _exit_status(retcode)
        return ret
    return True


def reboot(vm, force=False, key="uuid"):
    """
    Reboot a vm

    vm : string
        vm to be rebooted
    force : boolean
        force reboot of vm if true
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.reboot 186da9ab-7392-4f55-91a5-b8f1fe770543
        salt '*' vmadm.reboot 186da9ab-7392-4f55-91a5-b8f1fe770543 True
        salt '*' vmadm.reboot vm=nacl key=alias
        salt '*' vmadm.reboot vm=nina.example.org key=hostname
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup("{}={}".format(key, vm), one=True)
    if "Error" in vm:
        return vm
    # vmadm reboot <uuid> [-F]
    cmd = "vmadm reboot {force} {uuid}".format(force="-F" if force else "", uuid=vm)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def list_vms(search=None, sort=None, order="uuid,type,ram,state,alias", keyed=True):
    """
    Return a list of VMs

    search : string
        vmadm filter property
    sort : string
        vmadm sort (-s) property
    order : string
        vmadm order (-o) property -- Default: uuid,type,ram,state,alias
    keyed : boolean
        specified if the output should be an array (False) or dict (True)
            For a dict the key is the first item from the order parameter.
            Note: If key is not unique last vm wins.

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.list
        salt '*' vmadm.list order=alias,ram,cpu_cap sort=-ram,-cpu_cap
        salt '*' vmadm.list search='type=KVM'
    """
    ret = {}
    # vmadm list [-p] [-H] [-o field,...] [-s field,...] [field=value ...]
    cmd = "vmadm list -p -H {order} {sort} {search}".format(
        order="-o {}".format(order) if order else "",
        sort="-s {}".format(sort) if sort else "",
        search=search if search else "",
    )
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    result = OrderedDict() if keyed else []
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret

    fields = order.split(",")

    for vm in res["stdout"].splitlines():
        vm_data = OrderedDict()
        vm = vm.split(":")
        if keyed:
            for field in fields:
                if fields.index(field) == 0:
                    continue
                vm_data[field.strip()] = vm[fields.index(field)].strip()
            result[vm[0]] = vm_data
        else:
            if len(vm) > 1:
                for field in fields:
                    vm_data[field.strip()] = vm[fields.index(field)].strip()
            else:
                vm_data = vm[0]
            result.append(vm_data)
    return result


def lookup(search=None, order=None, one=False):
    """
    Return a list of VMs using lookup

    search : string
        vmadm filter property
    order : string
        vmadm order (-o) property -- Default: uuid,type,ram,state,alias
    one : boolean
        return only one result (vmadm's -1)

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.lookup search='state=running'
        salt '*' vmadm.lookup search='state=running' order=uuid,alias,hostname
        salt '*' vmadm.lookup search='alias=nacl' one=True
    """
    ret = {}
    # vmadm lookup [-j|-1] [-o field,...] [field=value ...]
    cmd = "vmadm lookup {one} {order} {search}".format(
        one="-1" if one else "-j",
        order="-o {}".format(order) if order else "",
        search=search if search else "",
    )
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    result = []
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret

    if one:
        result = res["stdout"]
    else:
        for vm in salt.utils.json.loads(res["stdout"]):
            result.append(vm)

    return result


def sysrq(vm, action="nmi", key="uuid"):
    """
    Send non-maskable interrupt to vm or capture a screenshot

    vm : string
        vm to be targeted
    action : string
        nmi or screenshot -- Default: nmi
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.sysrq 186da9ab-7392-4f55-91a5-b8f1fe770543 nmi
        salt '*' vmadm.sysrq 186da9ab-7392-4f55-91a5-b8f1fe770543 screenshot
        salt '*' vmadm.sysrq nacl nmi key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    if action not in ["nmi", "screenshot"]:
        ret["Error"] = "Action must be either nmi or screenshot"
        return ret
    vm = lookup("{}={}".format(key, vm), one=True)
    if "Error" in vm:
        return vm
    # vmadm sysrq <uuid> <nmi|screenshot>
    cmd = "vmadm sysrq {uuid} {action}".format(uuid=vm, action=action)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def delete(vm, key="uuid"):
    """
    Delete a vm

    vm : string
        vm to be deleted
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.delete 186da9ab-7392-4f55-91a5-b8f1fe770543
        salt '*' vmadm.delete nacl key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup("{}={}".format(key, vm), one=True)
    if "Error" in vm:
        return vm
    # vmadm delete <uuid>
    cmd = "vmadm delete {}".format(vm)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def get(vm, key="uuid"):
    """
    Output the JSON object describing a VM

    vm : string
        vm to be targeted
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.get 186da9ab-7392-4f55-91a5-b8f1fe770543
        salt '*' vmadm.get nacl key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup("{}={}".format(key, vm), one=True)
    if "Error" in vm:
        return vm
    # vmadm get <uuid>
    cmd = "vmadm get {}".format(vm)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return salt.utils.json.loads(res["stdout"])


def info(vm, info_type="all", key="uuid"):
    """
    Lookup info on running kvm

    vm : string
        vm to be targeted
    info_type : string [all|block|blockstats|chardev|cpus|kvm|pci|spice|version|vnc]
        info type to return
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.info 186da9ab-7392-4f55-91a5-b8f1fe770543
        salt '*' vmadm.info 186da9ab-7392-4f55-91a5-b8f1fe770543 vnc
        salt '*' vmadm.info nacl key=alias
        salt '*' vmadm.info nacl vnc key=alias
    """
    ret = {}
    if info_type not in [
        "all",
        "block",
        "blockstats",
        "chardev",
        "cpus",
        "kvm",
        "pci",
        "spice",
        "version",
        "vnc",
    ]:
        ret["Error"] = "Requested info_type is not available"
        return ret
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup("{}={}".format(key, vm), one=True)
    if "Error" in vm:
        return vm
    # vmadm info <uuid> [type,...]
    cmd = "vmadm info {uuid} {type}".format(uuid=vm, type=info_type)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return salt.utils.json.loads(res["stdout"])


def create_snapshot(vm, name, key="uuid"):
    """
    Create snapshot of a vm

    vm : string
        vm to be targeted
    name : string
        snapshot name
            The snapname must be 64 characters or less
            and must only contain alphanumeric characters and
            characters in the set [-_.:%] to comply with ZFS restrictions.
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.create_snapshot 186da9ab-7392-4f55-91a5-b8f1fe770543 baseline
        salt '*' vmadm.create_snapshot nacl baseline key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup("{}={}".format(key, vm), one=True)
    if "Error" in vm:
        return vm
    vmobj = get(vm)
    if "datasets" in vmobj:
        ret["Error"] = "VM cannot have datasets"
        return ret
    if vmobj["brand"] in ["kvm"]:
        ret["Error"] = "VM must be of type OS"
        return ret
    if vmobj["zone_state"] not in ["running"]:  # work around a vmadm bug
        ret["Error"] = "VM must be running to take a snapshot"
        return ret
    # vmadm create-snapshot <uuid> <snapname>
    cmd = "vmadm create-snapshot {uuid} {snapshot}".format(snapshot=name, uuid=vm)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def delete_snapshot(vm, name, key="uuid"):
    """
    Delete snapshot of a vm

    vm : string
        vm to be targeted
    name : string
        snapshot name
            The snapname must be 64 characters or less
            and must only contain alphanumeric characters and
            characters in the set [-_.:%] to comply with ZFS restrictions.
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.delete_snapshot 186da9ab-7392-4f55-91a5-b8f1fe770543 baseline
        salt '*' vmadm.delete_snapshot nacl baseline key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup("{}={}".format(key, vm), one=True)
    if "Error" in vm:
        return vm
    vmobj = get(vm)
    if "datasets" in vmobj:
        ret["Error"] = "VM cannot have datasets"
        return ret
    if vmobj["brand"] in ["kvm"]:
        ret["Error"] = "VM must be of type OS"
        return ret
    # vmadm delete-snapshot <uuid> <snapname>
    cmd = "vmadm delete-snapshot {uuid} {snapshot}".format(snapshot=name, uuid=vm)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def rollback_snapshot(vm, name, key="uuid"):
    """
    Rollback snapshot of a vm

    vm : string
        vm to be targeted
    name : string
        snapshot name
            The snapname must be 64 characters or less
            and must only contain alphanumeric characters and
            characters in the set [-_.:%] to comply with ZFS restrictions.
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.rollback_snapshot 186da9ab-7392-4f55-91a5-b8f1fe770543 baseline
        salt '*' vmadm.rollback_snapshot nacl baseline key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup("{}={}".format(key, vm), one=True)
    if "Error" in vm:
        return vm
    vmobj = get(vm)
    if "datasets" in vmobj:
        ret["Error"] = "VM cannot have datasets"
        return ret
    if vmobj["brand"] in ["kvm"]:
        ret["Error"] = "VM must be of type OS"
        return ret
    # vmadm rollback-snapshot <uuid> <snapname>
    cmd = "vmadm rollback-snapshot {uuid} {snapshot}".format(snapshot=name, uuid=vm)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def reprovision(vm, image, key="uuid"):
    """
    Reprovision a vm

    vm : string
        vm to be reprovisioned
    image : string
        uuid of new image
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.reprovision 186da9ab-7392-4f55-91a5-b8f1fe770543 c02a2044-c1bd-11e4-bd8c-dfc1db8b0182
        salt '*' vmadm.reprovision nacl c02a2044-c1bd-11e4-bd8c-dfc1db8b0182 key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    vm = lookup("{}={}".format(key, vm), one=True)
    if "Error" in vm:
        return vm
    if image not in __salt__["imgadm.list"]():
        ret["Error"] = "Image ({}) is not present on this host".format(image)
        return ret
    # vmadm reprovision <uuid> [-f <filename>]
    cmd = "echo {image} | vmadm reprovision {uuid}".format(
        uuid=salt.utils.stringutils.to_unicode(vm),
        image=_quote_args(salt.utils.json.dumps({"image_uuid": image})),
    )
    res = __salt__["cmd.run_all"](cmd, python_shell=True)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


def create(from_file=None, **kwargs):
    """
    Create a new vm

    from_file : string
        json file to create the vm from -- if present, all other options will be ignored
    kwargs : string|int|...
        options to set for the vm

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.create from_file=/tmp/new_vm.json
        salt '*' vmadm.create image_uuid='...' alias='...' nics='[{ "nic_tag": "admin", "ip": "198.51.100.123", ...}, {...}]' [...]
    """
    ret = {}
    # prepare vmcfg
    vmcfg = {}
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    for k, v in kwargs.items():
        vmcfg[k] = v

    if from_file:
        return _create_update_from_file("create", path=from_file)
    else:
        return _create_update_from_cfg("create", vmcfg=vmcfg)


def update(vm, from_file=None, key="uuid", **kwargs):
    """
    Update a new vm

    vm : string
        vm to be updated
    from_file : string
        json file to update the vm with -- if present, all other options will be ignored
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter
    kwargs : string|int|...
        options to update for the vm

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.update vm=186da9ab-7392-4f55-91a5-b8f1fe770543 from_file=/tmp/new_vm.json
        salt '*' vmadm.update vm=nacl key=alias from_file=/tmp/new_vm.json
        salt '*' vmadm.update vm=186da9ab-7392-4f55-91a5-b8f1fe770543 max_physical_memory=1024
    """
    ret = {}
    # prepare vmcfg
    vmcfg = {}
    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    for k, v in kwargs.items():
        vmcfg[k] = v

    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    uuid = lookup("{}={}".format(key, vm), one=True)
    if "Error" in uuid:
        return uuid

    if from_file:
        return _create_update_from_file("update", uuid, path=from_file)
    else:
        return _create_update_from_cfg("update", uuid, vmcfg=vmcfg)


def send(vm, target, key="uuid"):
    """
    Send a vm to a directory

    vm : string
        vm to be sent
    target : string
        target directory
    key : string [uuid|alias|hostname]
        value type of 'vm' parameter

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.send 186da9ab-7392-4f55-91a5-b8f1fe770543 /opt/backups
        salt '*' vmadm.send vm=nacl target=/opt/backups key=alias
    """
    ret = {}
    if key not in ["uuid", "alias", "hostname"]:
        ret["Error"] = "Key must be either uuid, alias or hostname"
        return ret
    if not os.path.isdir(target):
        ret["Error"] = "Target must be a directory or host"
        return ret
    vm = lookup("{}={}".format(key, vm), one=True)
    if "Error" in vm:
        return vm
    # vmadm send <uuid> [target]
    cmd = "vmadm send {uuid} > {target}".format(
        uuid=vm, target=os.path.join(target, "{}.vmdata".format(vm))
    )
    res = __salt__["cmd.run_all"](cmd, python_shell=True)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    vmobj = get(vm)
    if "datasets" not in vmobj:
        return True
    log.warning("one or more datasets detected, this is not supported!")
    log.warning("trying to zfs send datasets...")
    for dataset in vmobj["datasets"]:
        name = dataset.split("/")
        name = name[-1]
        cmd = "zfs send {dataset} > {target}".format(
            dataset=dataset,
            target=os.path.join(target, "{}-{}.zfsds".format(vm, name)),
        )
        res = __salt__["cmd.run_all"](cmd, python_shell=True)
        retcode = res["retcode"]
        if retcode != 0:
            ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
            return ret
    return True


def receive(uuid, source):
    """
    Receive a vm from a directory

    uuid : string
        uuid of vm to be received
    source : string
        source directory

    CLI Example:

    .. code-block:: bash

        salt '*' vmadm.receive 186da9ab-7392-4f55-91a5-b8f1fe770543 /opt/backups
    """
    ret = {}
    if not os.path.isdir(source):
        ret["Error"] = "Source must be a directory or host"
        return ret
    if not os.path.exists(os.path.join(source, "{}.vmdata".format(uuid))):
        ret["Error"] = "Unknow vm with uuid in {}".format(source)
        return ret
    # vmadm receive
    cmd = "vmadm receive < {source}".format(
        source=os.path.join(source, "{}.vmdata".format(uuid))
    )
    res = __salt__["cmd.run_all"](cmd, python_shell=True)
    retcode = res["retcode"]
    if retcode != 0 and not res["stderr"].endswith("datasets"):
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    vmobj = get(uuid)
    if "datasets" not in vmobj:
        return True
    log.warning("one or more datasets detected, this is not supported!")
    log.warning("trying to restore datasets, mountpoints will need to be set again...")
    for dataset in vmobj["datasets"]:
        name = dataset.split("/")
        name = name[-1]
        cmd = "zfs receive {dataset} < {source}".format(
            dataset=dataset,
            source=os.path.join(source, "{}-{}.zfsds".format(uuid, name)),
        )
        res = __salt__["cmd.run_all"](cmd, python_shell=True)
        retcode = res["retcode"]
        if retcode != 0:
            ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
            return ret
    cmd = "vmadm install {}".format(uuid)
    res = __salt__["cmd.run_all"](cmd, python_shell=True)
    retcode = res["retcode"]
    if retcode != 0 and not res["stderr"].endswith("datasets"):
        ret["Error"] = res["stderr"] if "stderr" in res else _exit_status(retcode)
        return ret
    return True


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



"""
    :codeauthor: :email:`David Homolka <david.homolka@ultimum.io>`
"""

import os

import salt.modules.restartcheck as restartcheck
import salt.utils.path
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import ANY, MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase


class RestartcheckTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.restartcheck
    """

    def setup_loader_modules(self):
        return {restartcheck: {}}

    def test_kernel_versions_debian(self):
        """
        Test kernel version debian
        """
        mock = MagicMock(return_value="  Installed: 4.9.82-1+deb9u3")
        with patch.dict(restartcheck.__grains__, {"os": "Debian"}):
            with patch.dict(restartcheck.__salt__, {"cmd.run": mock}):
                self.assertListEqual(
                    restartcheck._kernel_versions_debian(), ["4.9.82-1+deb9u3"]
                )

    def test_kernel_versions_ubuntu(self):
        """
        Test kernel version ubuntu
        """
        mock = MagicMock(return_value="  Installed: 4.10.0-42.46")
        with patch.dict(restartcheck.__grains__, {"os": "Ubuntu"}):
            with patch.dict(restartcheck.__salt__, {"cmd.run": mock}):
                self.assertListEqual(
                    restartcheck._kernel_versions_debian(),
                    [
                        "4.10.0-42.46",
                        "4.10.0-42-generic #46",
                        "4.10.0-42-lowlatency #46",
                    ],
                )

    def test_kernel_versions_redhat(self):
        """
        Test if it return a data structure of the current, in-memory rules
        """
        mock = MagicMock(
            return_value=(
                "kernel-3.10.0-862.el7.x86_64                  Thu Apr 5 00:40:00 2018"
            )
        )
        with patch.dict(restartcheck.__salt__, {"cmd.run": mock}):
            self.assertListEqual(
                restartcheck._kernel_versions_redhat(), ["3.10.0-862.el7.x86_64"]
            )

    def test_valid_deleted_file_deleted(self):
        """
        Test (deleted) file
        """
        self.assertTrue(restartcheck._valid_deleted_file("/usr/lib/test (deleted)"))

    def test_valid_deleted_file_psth_inode(self):
        """
        Test (path inode=1) file
        """
        self.assertTrue(
            restartcheck._valid_deleted_file("/usr/lib/test (path inode=1)")
        )

    def test_valid_deleted_file_var_log(self):
        """
        Test /var/log/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/var/log/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/var/log/test (deleted)"))
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/log/test (path inode=1)")
        )

    def test_valid_deleted_file_var_local_log(self):
        """
        Test /var/local/log/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/var/local/log/test"))
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/local/log/test (deleted)")
        )
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/local/log/test (path inode=1)")
        )

    def test_valid_deleted_file_var_run(self):
        """
        Test /var/run/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/var/run/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/var/run/test (deleted)"))
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/run/test (path inode=1)")
        )

    def test_valid_deleted_file_var_local_run(self):
        """
        Test /var/local/run/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/var/local/run/test"))
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/local/run/test (deleted)")
        )
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/local/run/test (path inode=1)")
        )

    def test_valid_deleted_file_tmp(self):
        """
        Test /tmp/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/tmp/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/tmp/test (deleted)"))
        self.assertFalse(restartcheck._valid_deleted_file("/tmp/test (path inode=1)"))

    def test_valid_deleted_file_dev_shm(self):
        """
        Test /dev/shm/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/dev/shm/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/dev/shm/test (deleted)"))
        self.assertFalse(
            restartcheck._valid_deleted_file("/dev/shm/test (path inode=1)")
        )

    def test_valid_deleted_file_run(self):
        """
        Test /run/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/run/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/run/test (deleted)"))
        self.assertFalse(restartcheck._valid_deleted_file("/run/test (path inode=1)"))

    def test_valid_deleted_file_drm(self):
        """
        Test /drm/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/drm/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/drm/test (deleted)"))
        self.assertFalse(restartcheck._valid_deleted_file("/drm/test (path inode=1)"))

    def test_valid_deleted_file_var_tmp(self):
        """
        Test /var/tmp/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/var/tmp/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/var/tmp/test (deleted)"))
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/tmp/test (path inode=1)")
        )

    def test_valid_deleted_file_var_local_tmp(self):
        """
        Test /var/local/tmp/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/var/local/tmp/test"))
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/local/tmp/test (deleted)")
        )
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/local/tmp/test (path inode=1)")
        )

    def test_valid_deleted_file_dev_zero(self):
        """
        Test /dev/zero/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/dev/zero/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/dev/zero/test (deleted)"))
        self.assertFalse(
            restartcheck._valid_deleted_file("/dev/zero/test (path inode=1)")
        )

    def test_valid_deleted_file_dev_pts(self):
        """
        Test /dev/pts/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/dev/pts/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/dev/pts/test (deleted)"))
        self.assertFalse(
            restartcheck._valid_deleted_file("/dev/pts/test (path inode=1)")
        )

    def test_valid_deleted_file_usr_lib_locale(self):
        """
        Test /usr/lib/locale/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/usr/lib/locale/test"))
        self.assertFalse(
            restartcheck._valid_deleted_file("/usr/lib/locale/test (deleted)")
        )
        self.assertFalse(
            restartcheck._valid_deleted_file("/usr/lib/locale/test (path inode=1)")
        )

    def test_valid_deleted_file_home(self):
        """
        Test /home/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/home/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/home/test (deleted)"))
        self.assertFalse(restartcheck._valid_deleted_file("/home/test (path inode=1)"))

    def test_valid_deleted_file_icon_theme_cache(self):
        """
        Test /test.icon-theme.cache
        """
        self.assertFalse(restartcheck._valid_deleted_file("/dev/test.icon-theme.cache"))
        self.assertFalse(
            restartcheck._valid_deleted_file("/dev/test.icon-theme.cache (deleted)")
        )
        self.assertFalse(
            restartcheck._valid_deleted_file(
                "/dev/test.icon-theme.cache (path inode=1)"
            )
        )

    def test_valid_deleted_file_var_cache_fontconfig(self):
        """
        Test /var/cache/fontconfig/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/var/cache/fontconfig/test"))
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/cache/fontconfig/test (deleted)")
        )
        self.assertFalse(
            restartcheck._valid_deleted_file(
                "/var/cache/fontconfig/test (path inode=1)"
            )
        )

    def test_valid_deleted_file_var_lib_nagios3_spool(self):
        """
        Test /var/lib/nagios3/spool/
        """
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/lib/nagios3/spool/test")
        )
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/lib/nagios3/spool/test (deleted)")
        )
        self.assertFalse(
            restartcheck._valid_deleted_file(
                "/var/lib/nagios3/spool/test (path inode=1)"
            )
        )

    def test_valid_deleted_file_var_lib_nagios3_spool_checkresults(self):
        """
        Test /var/lib/nagios3/spool/checkresults/
        """
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/lib/nagios3/spool/checkresults/test")
        )
        self.assertFalse(
            restartcheck._valid_deleted_file(
                "/var/lib/nagios3/spool/checkresults/test (deleted)"
            )
        )
        self.assertFalse(
            restartcheck._valid_deleted_file(
                "/var/lib/nagios3/spool/checkresults/test (path inode=1)"
            )
        )

    def test_valid_deleted_file_var_lib_postgresql(self):
        """
        Test /var/lib/postgresql/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/var/lib/postgresql/test"))
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/lib/postgresql/test (deleted)")
        )
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/lib/postgresql/test (path inode=1)")
        )

    def test_valid_deleted_file_var_lib_vdr(self):
        """
        Test /var/lib/vdr/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/var/lib/vdr/test"))
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/lib/vdr/test (deleted)")
        )
        self.assertFalse(
            restartcheck._valid_deleted_file("/var/lib/vdr/test (path inode=1)")
        )

    def test_valid_deleted_file_aio(self):
        """
        Test /[aio]/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/opt/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/opt/test (deleted)"))
        self.assertFalse(restartcheck._valid_deleted_file("/opt/test (path inode=1)"))
        self.assertFalse(restartcheck._valid_deleted_file("/apt/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/apt/test (deleted)"))
        self.assertFalse(restartcheck._valid_deleted_file("/apt/test (path inode=1)"))
        self.assertFalse(restartcheck._valid_deleted_file("/ipt/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/ipt/test (deleted)"))
        self.assertFalse(restartcheck._valid_deleted_file("/ipt/test (path inode=1)"))
        self.assertFalse(restartcheck._valid_deleted_file("/aio/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/aio/test (deleted)"))
        self.assertFalse(restartcheck._valid_deleted_file("/aio/test (path inode=1)"))

    def test_valid_deleted_file_sysv(self):
        """
        Test /SYSV/
        """
        self.assertFalse(restartcheck._valid_deleted_file("/SYSV/test"))
        self.assertFalse(restartcheck._valid_deleted_file("/SYSV/test (deleted)"))
        self.assertFalse(restartcheck._valid_deleted_file("/SYSV/test (path inode=1)"))

    def test_valid_command(self):
        """
        test for CVE-2020-28243
        """
        create_file = os.path.join(RUNTIME_VARS.TMP, "created_file")

        patch_kernel = patch(
            "salt.modules.restartcheck._kernel_versions_redhat",
            return_value=["3.10.0-1127.el7.x86_64"],
        )
        services = {
            "NetworkManager": {"ExecMainPID": 123},
            "auditd": {"ExecMainPID": 456},
            "crond": {"ExecMainPID": 789},
        }

        patch_salt = patch.dict(
            restartcheck.__salt__,
            {
                "cmd.run": MagicMock(
                    return_value="Linux localhost.localdomain 3.10.0-1127.el7.x86_64"
                ),
                "service.get_running": MagicMock(return_value=list(services.keys())),
                "service.show": MagicMock(side_effect=list(services.values())),
                "pkg.owner": MagicMock(return_value=""),
                "service.available": MagicMock(return_value=True),
            },
        )

        patch_deleted = patch(
            "salt.modules.restartcheck._deleted_files",
            MagicMock(
                return_value=[
                    (";touch {};".format(create_file), 123, "/root/ (deleted)")
                ]
            ),
        )

        patch_readlink = patch(
            "os.readlink", return_value="/root/;touch {};".format(create_file)
        )

        check_error = True
        if salt.utils.path.which("repoquery"):
            check_error = False

        patch_grains = patch.dict(restartcheck.__grains__, {"os_family": "RedHat"})
        with patch_kernel, patch_salt, patch_deleted, patch_readlink, patch_grains:
            if check_error:
                with self.assertRaises(FileNotFoundError):
                    restartcheck.restartcheck()
            else:
                ret = restartcheck.restartcheck()
                self.assertIn(
                    "Found 1 processes using old versions of upgraded files", ret
                )
            self.assertFalse(os.path.exists(create_file))

    def test_valid_command_b(self):
        """
        test for CVE-2020-28243
        """
        create_file = os.path.join(RUNTIME_VARS.TMP, "created_file")

        patch_kernel = patch(
            "salt.modules.restartcheck._kernel_versions_redhat",
            return_value=["3.10.0-1127.el7.x86_64"],
        )
        services = {
            "NetworkManager": {"ExecMainPID": 123},
            "auditd": {"ExecMainPID": 456},
            "crond": {"ExecMainPID": 789},
        }

        patch_salt = patch.dict(
            restartcheck.__salt__,
            {
                "cmd.run": MagicMock(
                    return_value="Linux localhost.localdomain 3.10.0-1127.el7.x86_64"
                ),
                "service.get_running": MagicMock(return_value=list(services.keys())),
                "service.show": MagicMock(side_effect=list(services.values())),
                "pkg.owner": MagicMock(return_value=""),
                "service.available": MagicMock(return_value=True),
            },
        )

        patch_deleted = patch(
            "salt.modules.restartcheck._deleted_files",
            MagicMock(return_value=[("--admindir tmp dpkg", 123, "/root/ (deleted)")]),
        )

        patch_readlink = patch("os.readlink", return_value="--admindir tmp dpkg")

        popen_mock = MagicMock()
        popen_mock.return_value.stdout.readline.side_effect = ["/usr/bin\n", ""]
        patch_popen = patch("subprocess.Popen", popen_mock)

        patch_grains = patch.dict(restartcheck.__grains__, {"os_family": "RedHat"})
        with patch_kernel, patch_salt, patch_deleted, patch_readlink, patch_grains, patch_popen:
            ret = restartcheck.restartcheck()
            self.assertIn("Found 1 processes using old versions of upgraded files", ret)
            popen_mock.assert_called_with(
                ["repoquery", "-l", "--admindir tmp dpkg"], stdout=ANY
            )



"""
    sphinxcontrib.httpdomain
    ~~~~~~~~~~~~~~~~~~~~~~~~

    The HTTP domain for documenting RESTful HTTP APIs.

    :copyright: Copyright 2011 by Hong Minhee
    :license: BSD, see LICENSE for details.

"""

import re

from docutils import nodes
from docutils.parsers.rst.roles import set_classes
from pygments.lexer import RegexLexer, bygroups
from pygments.lexers import get_lexer_by_name
from pygments.token import Keyword, Literal, Name, Number, Operator, Text
from pygments.util import ClassNotFound
from sphinx import addnodes
from sphinx.directives import ObjectDescription
from sphinx.domains import Domain, Index, ObjType
from sphinx.roles import XRefRole
from sphinx.util.docfields import GroupedField, TypedField
from sphinx.util.nodes import make_refnode


class DocRef:
    """Represents a link to an RFC which defines an HTTP method."""

    def __init__(self, base_url, anchor, section):
        """Stores the specified attributes which represent a URL which links to
        an RFC which defines an HTTP method.

        """
        self.base_url = base_url
        self.anchor = anchor
        self.section = section

    def __repr__(self):
        """Returns the URL which this object represents, which points to the
        location of the RFC which defines some HTTP method.

        """
        return "{}#{}{}".format(self.base_url, self.anchor, self.section)


#: The URL of the HTTP/1.1 RFC which defines the HTTP methods OPTIONS, GET,
#: HEAD, POST, PUT, DELETE, TRACE, and CONNECT.
RFC2616 = "http://www.w3.org/Protocols/rfc2616/rfc2616-sec9.html"

#: The name to use for section anchors in RFC2616.
RFC2616ANCHOR = "sec"

#: The URL of the RFC which defines the HTTP PATCH method.
RFC5789 = "http://tools.ietf.org/html/rfc5789"

#: The name to use for section anchors in RFC5789.
RFC5789ANCHOR = "section-"

#: Mapping from lowercase HTTP method name to :class:`DocRef` object which
#: maintains the URL which points to the section of the RFC which defines that
#: HTTP method.
DOCREFS = {
    "patch": DocRef(RFC5789, RFC5789ANCHOR, 2),
    "options": DocRef(RFC2616, RFC2616ANCHOR, 9.2),
    "get": DocRef(RFC2616, RFC2616ANCHOR, 9.3),
    "head": DocRef(RFC2616, RFC2616ANCHOR, 9.4),
    "post": DocRef(RFC2616, RFC2616ANCHOR, 9.5),
    "put": DocRef(RFC2616, RFC2616ANCHOR, 9.6),
    "delete": DocRef(RFC2616, RFC2616ANCHOR, 9.7),
    "trace": DocRef(RFC2616, RFC2616ANCHOR, 9.8),
    "connect": DocRef(RFC2616, RFC2616ANCHOR, 9.9),
}

HTTP_STATUS_CODES = {
    100: "Continue",
    101: "Switching Protocols",
    102: "Processing",
    200: "OK",
    201: "Created",
    202: "Accepted",
    203: "Non Authoritative Information",
    204: "No Content",
    205: "Reset Content",
    206: "Partial Content",
    207: "Multi Status",
    226: "IM Used",  # see RFC 3229
    300: "Multiple Choices",
    301: "Moved Permanently",
    302: "Found",
    303: "See Other",
    304: "Not Modified",
    305: "Use Proxy",
    307: "Temporary Redirect",
    400: "Bad Request",
    401: "Unauthorized",
    402: "Payment Required",  # unused
    403: "Forbidden",
    404: "Not Found",
    405: "Method Not Allowed",
    406: "Not Acceptable",
    407: "Proxy Authentication Required",
    408: "Request Timeout",
    409: "Conflict",
    410: "Gone",
    411: "Length Required",
    412: "Precondition Failed",
    413: "Request Entity Too Large",
    414: "Request URI Too Long",
    415: "Unsupported Media Type",
    416: "Requested Range Not Satisfiable",
    417: "Expectation Failed",
    418: "I'm a teapot",  # see RFC 2324
    422: "Unprocessable Entity",
    423: "Locked",
    424: "Failed Dependency",
    426: "Upgrade Required",
    449: "Retry With",  # proprietary MS extension
    500: "Internal Server Error",
    501: "Not Implemented",
    502: "Bad Gateway",
    503: "Service Unavailable",
    504: "Gateway Timeout",
    505: "HTTP Version Not Supported",
    507: "Insufficient Storage",
    510: "Not Extended",
}

http_sig_param_re = re.compile(
    r"\((?:(?P<type>[^:)]+):)?(?P<name>[\w_]+)\)", re.VERBOSE
)


def http_resource_anchor(method, path):
    path = re.sub(r"[<>:/]", "-", path)
    return method.lower() + "-" + path


class HTTPResource(ObjectDescription):

    doc_field_types = [
        TypedField(
            "parameter",
            label="Parameters",
            names=("param", "parameter", "arg", "argument"),
            typerolename="obj",
            typenames=("paramtype", "type"),
        ),
        TypedField(
            "jsonparameter",
            label="JSON Parameters",
            names=("jsonparameter", "jsonparam", "json"),
            typerolename="obj",
            typenames=("jsonparamtype", "jsontype"),
        ),
        TypedField(
            "queryparameter",
            label="Query Parameters",
            names=("queryparameter", "queryparam", "qparam", "query"),
            typerolename="obj",
            typenames=("queryparamtype", "querytype", "qtype"),
        ),
        GroupedField(
            "formparameter",
            label="Form Parameters",
            names=("formparameter", "formparam", "fparam", "form"),
        ),
        GroupedField(
            "requestheader",
            label="Request Headers",
            rolename="mailheader",
            names=("reqheader", "requestheader"),
        ),
        GroupedField(
            "responseheader",
            label="Response Headers",
            rolename="mailheader",
            names=("resheader", "responseheader"),
        ),
        GroupedField(
            "statuscode",
            label="Status Codes",
            rolename="statuscode",
            names=("statuscode", "status", "code"),
        ),
    ]

    method = NotImplemented

    def handle_signature(self, sig, signode):
        method = self.method.upper() + " "
        signode += addnodes.desc_name(method, method)
        offset = 0
        for match in http_sig_param_re.finditer(sig):
            path = sig[offset : match.start()]
            signode += addnodes.desc_name(path, path)
            params = addnodes.desc_parameterlist()
            typ = match.group("type")
            if typ:
                typ = typ + ": "
                params += addnodes.desc_annotation(typ, typ)
            name = match.group("name")
            params += addnodes.desc_parameter(name, name)
            signode += params
            offset = match.end()
        if offset < len(sig):
            path = sig[offset : len(sig)]
            signode += addnodes.desc_name(path, path)
        fullname = self.method.upper() + " " + path
        signode["method"] = self.method
        signode["path"] = sig
        signode["fullname"] = fullname
        return (fullname, self.method, sig)

    def needs_arglist(self):
        return False

    def add_target_and_index(self, name_cls, sig, signode):
        signode["ids"].append(http_resource_anchor(*name_cls[1:]))
        self.env.domaindata["http"][self.method][sig] = (self.env.docname, "")

    def get_index_text(self, modname, name):
        return ""


class HTTPOptions(HTTPResource):

    method = "options"


class HTTPHead(HTTPResource):

    method = "head"


class HTTPPatch(HTTPResource):

    method = "patch"


class HTTPPost(HTTPResource):

    method = "post"


class HTTPGet(HTTPResource):

    method = "get"


class HTTPPut(HTTPResource):

    method = "put"


class HTTPDelete(HTTPResource):

    method = "delete"


class HTTPTrace(HTTPResource):

    method = "trace"


def http_statuscode_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    if text.isdigit():
        code = int(text)
        try:
            status = HTTP_STATUS_CODES[code]
        except KeyError:
            msg = inliner.reporter.error(
                "%d is invalid HTTP status code" % code, lineno=lineno
            )
            prb = inliner.problematic(rawtext, rawtext, msg)
            return [prb], [msg]
    else:
        try:
            code, status = re.split(r"\s", text.strip(), 1)
            code = int(code)
        except ValueError:
            msg = inliner.reporter.error(
                "HTTP status code must be an integer (e.g. `200`) or "
                "start with an integer (e.g. `200 OK`); %r is invalid" % text,
                line=lineno,
            )
            prb = inliner.problematic(rawtext, rawtext, msg)
            return [prb], [msg]
    nodes.reference(rawtext)
    if code == 226:
        url = "http://www.ietf.org/rfc/rfc3229.txt"
    if code == 418:
        url = "http://www.ietf.org/rfc/rfc2324.txt"
    if code == 449:
        url = "http://msdn.microsoft.com/en-us/library/dd891478(v=prot.10).aspx"
    elif code in HTTP_STATUS_CODES:
        url = (
            "http://www.w3.org/Protocols/rfc2616/rfc2616-sec10.html#sec10."
            + "%d.%d" % (code // 100, 1 + code % 100)
        )
    else:
        url = ""
    set_classes(options)
    node = nodes.reference(rawtext, "%d %s" % (code, status), refuri=url, **options)
    return [node], []


def http_method_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    method = str(text).lower()
    if method not in DOCREFS:
        msg = inliner.reporter.error(
            "%s is not valid HTTP method" % method, lineno=lineno
        )
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    url = str(DOCREFS[method])
    node = nodes.reference(rawtext, method.upper(), refuri=url, **options)
    return [node], []


class HTTPXRefRole(XRefRole):
    def __init__(self, method, **kwargs):
        XRefRole.__init__(self, **kwargs)
        self.method = method

    def process_link(self, env, refnode, has_explicit_title, title, target):
        if not target.startswith("/"):
            pass
        if not has_explicit_title:
            title = self.method.upper() + " " + title
        return title, target


class HTTPIndex(Index):

    name = "routingtable"
    localname = "HTTP Routing Table"
    shortname = "routing table"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.ignore = [
            [l for l in x.split("/") if l]
            for x in self.domain.env.config["http_index_ignore_prefixes"]
        ]
        self.ignore.sort(key=lambda x: -len(x))

    def grouping_prefix(self, path):
        letters = [x for x in path.split("/") if x]
        for prefix in self.ignore:
            if letters[: len(prefix)] == prefix:
                return "/" + "/".join(letters[: len(prefix) + 1])
        return "/{}".format(letters[0] if letters else "")

    def generate(self, docnames=None):
        content = {}
        items = (
            (method, path, info)
            for method, routes in self.domain.routes.items()
            for path, info in routes.items()
        )
        items = sorted(items, key=lambda item: item[1])
        for method, path, info in items:
            entries = content.setdefault(self.grouping_prefix(path), [])
            entries.append(
                [
                    method.upper() + " " + path,
                    0,
                    info[0],
                    http_resource_anchor(method, path),
                    "",
                    "",
                    info[1],
                ]
            )
        content = sorted(content.items(), key=lambda k: k[0])
        return (content, True)


class HTTPDomain(Domain):
    """HTTP domain."""

    name = "http"
    label = "HTTP"

    object_types = {
        "options": ObjType("options", "options", "obj"),
        "head": ObjType("head", "head", "obj"),
        "post": ObjType("post", "post", "obj"),
        "get": ObjType("get", "get", "obj"),
        "put": ObjType("put", "put", "obj"),
        "patch": ObjType("patch", "patch", "obj"),
        "delete": ObjType("delete", "delete", "obj"),
        "trace": ObjType("trace", "trace", "obj"),
    }

    directives = {
        "options": HTTPOptions,
        "head": HTTPHead,
        "post": HTTPPost,
        "get": HTTPGet,
        "put": HTTPPut,
        "patch": HTTPPatch,
        "delete": HTTPDelete,
        "trace": HTTPTrace,
    }

    roles = {
        "options": HTTPXRefRole("options"),
        "head": HTTPXRefRole("head"),
        "post": HTTPXRefRole("post"),
        "get": HTTPXRefRole("get"),
        "put": HTTPXRefRole("put"),
        "patch": HTTPXRefRole("patch"),
        "delete": HTTPXRefRole("delete"),
        "trace": HTTPXRefRole("trace"),
        "statuscode": http_statuscode_role,
        "method": http_method_role,
    }

    initial_data = {
        "options": {},  # path: (docname, synopsis)
        "head": {},
        "post": {},
        "get": {},
        "put": {},
        "patch": {},
        "delete": {},
        "trace": {},
    }

    # indices = [HTTPIndex]
    indices = []

    @property
    def routes(self):
        return {key: self.data[key] for key in self.object_types}

    def clear_doc(self, docname):
        for typ, routes in self.routes.items():
            for path, info in list(routes.items()):
                if info[0] == docname:
                    del routes[path]

    def resolve_xref(self, env, fromdocname, builder, typ, target, node, contnode):
        try:
            info = self.data[str(typ)][target]
        except KeyError:
            return
        else:
            anchor = http_resource_anchor(typ, target)
            title = typ.upper() + " " + target
            return make_refnode(builder, fromdocname, info[0], anchor, contnode, title)

    def get_objects(self):
        for method, routes in self.routes.items():
            for path, info in routes.items():
                anchor = http_resource_anchor(method, path)
                yield (path, path, method, info[0], anchor, 1)


class HTTPLexer(RegexLexer):
    """Lexer for HTTP sessions."""

    name = "HTTP"
    aliases = ["http"]

    flags = re.DOTALL

    def header_callback(self, match):
        if match.group(1).lower() == "content-type":
            content_type = match.group(5).strip()
            if ";" in content_type:
                content_type = content_type[: content_type.find(";")].strip()
            self.content_type = content_type
        yield match.start(1), Name.Attribute, match.group(1)
        yield match.start(2), Text, match.group(2)
        yield match.start(3), Operator, match.group(3)
        yield match.start(4), Text, match.group(4)
        yield match.start(5), Literal, match.group(5)
        yield match.start(6), Text, match.group(6)

    def continuous_header_callback(self, match):
        yield match.start(1), Text, match.group(1)
        yield match.start(2), Literal, match.group(2)
        yield match.start(3), Text, match.group(3)

    def content_callback(self, match):
        content_type = getattr(self, "content_type", None)
        content = match.group()
        offset = match.start()
        if content_type:
            from pygments.lexers import get_lexer_for_mimetype

            try:
                lexer = get_lexer_for_mimetype(content_type)
            except ClassNotFound:
                pass
            else:
                for idx, token, value in lexer.get_tokens_unprocessed(content):
                    yield offset + idx, token, value
                return
        yield offset, Text, content

    tokens = {
        "root": [
            (
                r"(GET|POST|PUT|PATCH|DELETE|HEAD|OPTIONS|TRACE)( +)([^ ]+)( +)"
                r"(HTTPS?)(/)(1\.[01])(\r?\n|$)",
                bygroups(
                    Name.Function,
                    Text,
                    Name.Namespace,
                    Text,
                    Keyword.Reserved,
                    Operator,
                    Number,
                    Text,
                ),
                "headers",
            ),
            (
                r"(HTTPS?)(/)(1\.[01])( +)(\d{3})( +)([^\r\n]+)(\r?\n|$)",
                bygroups(
                    Keyword.Reserved,
                    Operator,
                    Number,
                    Text,
                    Number,
                    Text,
                    Name.Exception,
                    Text,
                ),
                "headers",
            ),
        ],
        "headers": [
            (r"([^\s:]+)( *)(:)( *)([^\r\n]+)(\r?\n|$)", header_callback),
            (r"([\t ]+)([^\r\n]+)(\r?\n|$)", continuous_header_callback),
            (r"\r?\n", Text, "content"),
        ],
        "content": [(r".+", content_callback)],
    }


def setup(app):
    app.add_domain(HTTPDomain)
    try:
        get_lexer_by_name("http")
    except ClassNotFound:
        app.add_lexer("http", HTTPLexer())
    app.add_config_value("http_index_ignore_prefixes", [], None)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import socket

import salt.modules.win_network as win_network
import salt.utils.network
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, Mock, patch
from tests.support.unit import TestCase, skipIf

try:
    import wmi

    HAS_WMI = True
except ImportError:
    HAS_WMI = False


class Mockwmi:
    """
    Mock wmi class
    """

    NetConnectionID = "Ethernet"

    def __init__(self):
        pass


class Mockwinapi:
    """
    Mock winapi class
    """

    def __init__(self):
        pass

    class winapi:
        """
        Mock winapi class
        """

        def __init__(self):
            pass

        class Com:
            """
            Mock Com method
            """

            def __enter__(self):
                return self

            def __exit__(self, *exc_info):
                return False


class WinNetworkTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.win_network
    """

    def setup_loader_modules(self):
        self.WMI = Mock()
        self.addCleanup(delattr, self, "WMI")
        return {win_network: {}}

    # 'ping' function tests: 1

    def test_ping(self):
        """
        Test if it performs a ping to a host.
        """
        mock = MagicMock(return_value=True)
        with patch.dict(win_network.__salt__, {"cmd.run": mock}):
            self.assertTrue(win_network.ping("127.0.0.1"))

    # 'netstat' function tests: 1

    def test_netstat(self):
        """
        Test if it return information on open ports and states
        """
        ret = (
            "  Proto  Local Address    Foreign Address    State    PID\n"
            "  TCP    127.0.0.1:1434    0.0.0.0:0    LISTENING    1728\n"
            "  UDP    127.0.0.1:1900    *:*        4240"
        )
        mock = MagicMock(return_value=ret)
        with patch.dict(win_network.__salt__, {"cmd.run": mock}):
            self.assertListEqual(
                win_network.netstat(),
                [
                    {
                        "local-address": "127.0.0.1:1434",
                        "program": "1728",
                        "proto": "TCP",
                        "remote-address": "0.0.0.0:0",
                        "state": "LISTENING",
                    },
                    {
                        "local-address": "127.0.0.1:1900",
                        "program": "4240",
                        "proto": "UDP",
                        "remote-address": "*:*",
                        "state": None,
                    },
                ],
            )

    # 'traceroute' function tests: 1

    def test_traceroute(self):
        """
        Test if it performs a traceroute to a 3rd party host
        """
        ret = (
            "  1     1 ms    <1 ms    <1 ms  172.27.104.1\n"
            "  2     1 ms    <1 ms     1 ms  121.242.35.1.s[121.242.35.1]\n"
            "  3     3 ms     2 ms     2 ms  121.242.4.53.s[121.242.4.53]\n"
        )
        mock = MagicMock(return_value=ret)
        with patch.dict(win_network.__salt__, {"cmd.run": mock}):
            self.assertListEqual(
                win_network.traceroute("google.com"),
                [
                    {
                        "count": "1",
                        "hostname": None,
                        "ip": "172.27.104.1",
                        "ms1": "1",
                        "ms2": "<1",
                        "ms3": "<1",
                    },
                    {
                        "count": "2",
                        "hostname": None,
                        "ip": "121.242.35.1.s[121.242.35.1]",
                        "ms1": "1",
                        "ms2": "<1",
                        "ms3": "1",
                    },
                    {
                        "count": "3",
                        "hostname": None,
                        "ip": "121.242.4.53.s[121.242.4.53]",
                        "ms1": "3",
                        "ms2": "2",
                        "ms3": "2",
                    },
                ],
            )

    # 'nslookup' function tests: 1

    def test_nslookup(self):
        """
        Test if it query DNS for information about a domain or ip address
        """
        ret = (
            "Server:  ct-dc-3-2.cybage.com\n"
            "Address:  172.27.172.12\n"
            "Non-authoritative answer:\n"
            "Name:    google.com\n"
            "Addresses:  2404:6800:4007:806::200e\n"
            "216.58.196.110\n"
        )
        mock = MagicMock(return_value=ret)
        with patch.dict(win_network.__salt__, {"cmd.run": mock}):
            self.assertListEqual(
                win_network.nslookup("google.com"),
                [
                    {"Server": "ct-dc-3-2.cybage.com"},
                    {"Address": "172.27.172.12"},
                    {"Name": "google.com"},
                    {"Addresses": ["2404:6800:4007:806::200e", "216.58.196.110"]},
                ],
            )

    # 'dig' function tests: 1

    def test_dig(self):
        """
        Test if it performs a DNS lookup with dig
        """
        mock = MagicMock(return_value=True)
        with patch.dict(win_network.__salt__, {"cmd.run": mock}):
            self.assertTrue(win_network.dig("google.com"))

    # 'interfaces_names' function tests: 1

    @skipIf(not HAS_WMI, "WMI only available on Windows")
    def test_interfaces_names(self):
        """
        Test if it return a list of all the interfaces names
        """
        self.WMI.Win32_NetworkAdapter = MagicMock(return_value=Mockwmi)
        with patch("salt.utils.winapi.Com", MagicMock()), patch.object(
            self.WMI, "Win32_NetworkAdapter", return_value=[Mockwmi()]
        ), patch("salt.utils", Mockwinapi), patch.object(
            wmi, "WMI", Mock(return_value=self.WMI)
        ):
            self.assertListEqual(win_network.interfaces_names(), ["Ethernet"])

    # 'interfaces' function tests: 1

    def test_interfaces(self):
        """
        Test if it return information about all the interfaces on the minion
        """
        with patch.object(
            salt.utils.network, "win_interfaces", MagicMock(return_value=True)
        ):
            self.assertTrue(win_network.interfaces())

    # 'hw_addr' function tests: 1

    def test_hw_addr(self):
        """
        Test if it return the hardware address (a.k.a. MAC address)
        for a given interface
        """
        with patch.object(
            salt.utils.network, "hw_addr", MagicMock(return_value="Ethernet")
        ):
            self.assertEqual(win_network.hw_addr("Ethernet"), "Ethernet")

    # 'subnets' function tests: 1

    def test_subnets(self):
        """
        Test if it returns a list of subnets to which the host belongs
        """
        with patch.object(
            salt.utils.network, "subnets", MagicMock(return_value="10.1.1.0/24")
        ):
            self.assertEqual(win_network.subnets(), "10.1.1.0/24")

    # 'in_subnet' function tests: 1

    def test_in_subnet(self):
        """
        Test if it returns True if host is within specified subnet,
        otherwise False
        """
        with patch.object(
            salt.utils.network, "in_subnet", MagicMock(return_value=True)
        ):
            self.assertTrue(win_network.in_subnet("10.1.1.0/16"))

    # 'get_route' function tests: 1

    def test_get_route(self):
        """
        Test if it return information on open ports and states
        """
        ret = (
            "\n\n"
            "IPAddress         : 10.0.0.15\n"
            "InterfaceIndex    : 3\n"
            "InterfaceAlias    : Wi-Fi\n"
            "AddressFamily     : IPv4\n"
            "Type              : Unicast\n"
            "PrefixLength      : 24\n"
            "PrefixOrigin      : Dhcp\n"
            "SuffixOrigin      : Dhcp\n"
            "AddressState      : Preferred\n"
            "ValidLifetime     : 6.17:52:39\n"
            "PreferredLifetime : 6.17:52:39\n"
            "SkipAsSource      : False\n"
            "PolicyStore       : ActiveStore\n"
            "\n\n"
            "Caption            :\n"
            "Description        :\n"
            "ElementName        :\n"
            "InstanceID         : :8:8:8:9:55=55;:8;8;:8;55;\n"
            "AdminDistance      :\n"
            "DestinationAddress :\n"
            "IsStatic           :\n"
            "RouteMetric        : 0\n"
            "TypeOfRoute        : 3\n"
            "AddressFamily      : IPv4\n"
            "CompartmentId      : 1\n"
            "DestinationPrefix  : 0.0.0.0/0\n"
            "InterfaceAlias     : Wi-Fi\n"
            "InterfaceIndex     : 3\n"
            "NextHop            : 10.0.0.1\n"
            "PreferredLifetime  : 6.23:14:43\n"
            "Protocol           : NetMgmt\n"
            "Publish            : No\n"
            "Store              : ActiveStore\n"
            "ValidLifetime      : 6.23:14:43\n"
            "PSComputerName     :\n"
            "ifIndex            : 3"
        )
        mock = MagicMock(return_value=ret)
        with patch.dict(win_network.__salt__, {"cmd.run": mock}):
            self.assertDictEqual(
                win_network.get_route("192.0.0.8"),
                {
                    "destination": "192.0.0.8",
                    "gateway": "10.0.0.1",
                    "interface": "Wi-Fi",
                    "source": "10.0.0.15",
                },
            )

    def test_connect_53371(self):
        """
        Test that UnboundLocalError is not thrown on socket.gaierror
        as reported in #53371
        """
        with patch(
            "socket.getaddrinfo",
            autospec=True,
            side_effect=socket.gaierror("[Errno 11004] getaddrinfo failed"),
        ):
            rtn = win_network.connect("test-server", 80)
            assert rtn
            assert not rtn["result"]
            assert (
                rtn["comment"]
                == "Unable to connect to test-server (unknown) on tcp port 80"
            )



import pytest
import salt.states.win_license as win_license
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {win_license: {}}


def test_activate():
    """
    Test activating the given product key
    """
    expected = {
        "changes": {},
        "comment": "Windows is now activated.",
        "name": "AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE",
        "result": True,
    }

    info = {
        "description": "Prof",
        "licensed": False,
        "name": "Win7",
        "partial_key": "XXXXX",
    }

    info_mock = MagicMock(return_value=info)
    install_mock = MagicMock(return_value="Installed successfully")
    activate_mock = MagicMock(return_value="Activated successfully")
    with patch.dict(
        win_license.__salt__,
        {
            "license.info": info_mock,
            "license.install": install_mock,
            "license.activate": activate_mock,
        },
    ):
        out = win_license.activate("AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE")
        info_mock.assert_called_once_with()
        install_mock.assert_called_once_with("AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE")
        activate_mock.assert_called_once_with()
        assert out == expected


def test_installed_not_activated():
    """
    Test activating the given product key when the key is installed but not activated
    """
    expected = {
        "changes": {},
        "comment": "Windows is now activated.",
        "name": "AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE",
        "result": True,
    }

    info = {
        "description": "Prof",
        "licensed": False,
        "name": "Win7",
        "partial_key": "ABCDE",
    }

    info_mock = MagicMock(return_value=info)
    install_mock = MagicMock(return_value="Installed successfully")
    activate_mock = MagicMock(return_value="Activated successfully")
    with patch.dict(
        win_license.__salt__,
        {
            "license.info": info_mock,
            "license.install": install_mock,
            "license.activate": activate_mock,
        },
    ):
        out = win_license.activate("AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE")
        info_mock.assert_called_once_with()
        assert not install_mock.called
        activate_mock.assert_called_once_with()
        assert out == expected


def test_installed_activated():
    """
    Test activating the given product key when its already activated
    """
    expected = {
        "changes": {},
        "comment": "Windows is already activated.",
        "name": "AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE",
        "result": True,
    }

    info = {
        "description": "Prof",
        "licensed": True,
        "name": "Win7",
        "partial_key": "ABCDE",
    }

    info_mock = MagicMock(return_value=info)
    install_mock = MagicMock(return_value="Installed successfully")
    activate_mock = MagicMock(return_value="Activated successfully")
    with patch.dict(
        win_license.__salt__,
        {
            "license.info": info_mock,
            "license.install": install_mock,
            "license.activate": activate_mock,
        },
    ):
        out = win_license.activate("AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE")
        info_mock.assert_called_once_with()
        assert not install_mock.called
        assert not activate_mock.called
        assert out == expected


def test_installed_install_fail():
    """
    Test activating the given product key when the install fails
    """
    expected = {
        "changes": {},
        "comment": "Unable to install the given product key is it valid?",
        "name": "AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE",
        "result": False,
    }

    info = {
        "description": "Prof",
        "licensed": False,
        "name": "Win7",
        "partial_key": "12345",
    }

    info_mock = MagicMock(return_value=info)
    install_mock = MagicMock(return_value="Failed")
    activate_mock = MagicMock()
    with patch.dict(
        win_license.__salt__,
        {
            "license.info": info_mock,
            "license.install": install_mock,
            "license.activate": activate_mock,
        },
    ):
        out = win_license.activate("AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE")
        info_mock.assert_called_once_with()
        install_mock.assert_called_once_with("AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE")
        assert not activate_mock.called
        assert out == expected


def test_installed_activate_fail():
    """
    Test activating the given product key when the install fails
    """
    expected = {
        "changes": {},
        "comment": "Unable to activate the given product key.",
        "name": "AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE",
        "result": False,
    }

    info = {
        "description": "Prof",
        "licensed": False,
        "name": "Win7",
        "partial_key": "ABCDE",
    }

    info_mock = MagicMock(return_value=info)
    install_mock = MagicMock()
    activate_mock = MagicMock(return_value="Failed to activate")
    with patch.dict(
        win_license.__salt__,
        {
            "license.info": info_mock,
            "license.install": install_mock,
            "license.activate": activate_mock,
        },
    ):
        out = win_license.activate("AAAAA-AAAAA-AAAAA-AAAA-AAAAA-ABCDE")
        info_mock.assert_called_once_with()
        assert not install_mock.called
        activate_mock.assert_called_once_with()
        assert out == expected



import salt.fileserver.svnfs as svnfs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import patch
from tests.support.unit import TestCase


class SvnfsFileTest(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {svnfs: {}}

    def test_env_is_exposed(self):
        """
        test _env_is_exposed method when
        base is in whitelist
        """
        with patch.dict(
            svnfs.__opts__,
            {"svnfs_saltenv_whitelist": "base", "svnfs_saltenv_blacklist": ""},
        ):
            assert svnfs._env_is_exposed("base")

    def test_env_is_exposed_blacklist(self):
        """
        test _env_is_exposed method when
        base is in blacklist
        """
        with patch.dict(
            svnfs.__opts__,
            {"svnfs_saltenv_whitelist": "", "svnfs_saltenv_blacklist": "base"},
        ):
            assert not svnfs._env_is_exposed("base")



"""
    :codeauthor: :email:`Jorge Schrauwen <sjorge@blackdot.be>`
"""

import textwrap

import salt.grains.smartos as smartos
from tests.support.mock import MagicMock, Mock, mock_open, patch
from tests.support.unit import TestCase


class SmartOSGrainsTestCase(TestCase):
    """
    Test cases for smartos grains
    """

    def test_smartos_computenode_data(self):
        """
        Get a tally of running/stopped zones
        Output used form a test host with one running
        and one stopped of each vm type.
        """
        grains_exp_res = {
            "computenode_sdc_version": "7.0",
            "computenode_vm_capable": True,
            "computenode_vm_hw_virt": "vmx",
            "computenode_vms_running": 3,
            "computenode_vms_stopped": 3,
            "computenode_vms_total": 6,
            "computenode_vms_type": {"KVM": 2, "LX": 2, "OS": 2},
            "manufacturer": "Supermicro",
            "productname": "X8STi",
            "uuid": "534d4349-0002-2790-2500-2790250054c5",
        }

        cmd_mock = Mock(
            side_effect=[
                textwrap.dedent(
                    """\
                99e40ee7-a8f9-4b57-9225-e7bd19f64b07:test_hvm1:running:BHYV
                cde351a9-e23d-6856-e268-fff10fe603dc:test_hvm2:stopped:BHYV
                99e40ee7-a8f9-4b57-9225-e7bd19f64b07:test_hvm3:running:KVM
                cde351a9-e23d-6856-e268-fff10fe603dc:test_hvm4:stopped:KVM
                179b50ca-8a4d-4f28-bb08-54b2cd350aa5:test_zone1:running:OS
                42846fbc-c48a-6390-fd85-d7ac6a76464c:test_zone2:stopped:OS
                4fd2d7a4-38c4-4068-a2c8-74124364a109:test_zone3:running:LX
                717abe34-e7b9-4387-820e-0bb041173563:test_zone4:stopped:LX"""
                ),
                textwrap.dedent(
                    """\
                {
                  "Live Image": "20181011T004530Z",
                  "System Type": "SunOS",
                  "Boot Time": "1562528522",
                  "SDC Version": "7.0",
                  "Manufacturer": "Supermicro",
                  "Product": "X8STi",
                  "Serial Number": "1234567890",
                  "SKU Number": "To Be Filled By O.E.M.",
                  "HW Version": "1234567890",
                  "HW Family": "High-End Desktop",
                  "Setup": "false",
                  "VM Capable": true,
                  "Bhyve Capable": false,
                  "Bhyve Max Vcpus": 0,
                  "HVM API": false,
                  "CPU Type": "Intel(R) Xeon(R) CPU W3520 @ 2.67GHz",
                  "CPU Virtualization": "vmx",
                  "CPU Physical Cores": 1,
                  "Admin NIC Tag": "",
                  "UUID": "534d4349-0002-2790-2500-2790250054c5",
                  "Hostname": "sdc",
                  "CPU Total Cores": 8,
                  "MiB of Memory": "16375",
                  "Zpool": "zones",
                  "Zpool Disks": "c1t0d0,c1t1d0",
                  "Zpool Profile": "mirror",
                  "Zpool Creation": 1406392163,
                  "Zpool Size in GiB": 1797,
                  "Disks": {
                    "c1t0d0": {"Size in GB": 2000},
                    "c1t1d0": {"Size in GB": 2000}
                  },
                  "Boot Parameters": {
                    "smartos": "true",
                    "console": "text",
                    "boot_args": "",
                    "bootargs": ""
                  },
                  "Network Interfaces": {
                    "e1000g0": {"MAC Address": "00:00:00:00:00:01", "ip4addr": "123.123.123.123", "Link Status": "up", "NIC Names": ["admin"]},
                    "e1000g1": {"MAC Address": "00:00:00:00:00:05", "ip4addr": "", "Link Status": "down", "NIC Names": []}
                  },
                  "Virtual Network Interfaces": {
                  },
                  "Link Aggregations": {
                  }
                }"""
                ),
            ]
        )
        with patch.dict(smartos.__salt__, {"cmd.run": cmd_mock}):
            grains_res = smartos._smartos_computenode_data()
            self.assertEqual(grains_exp_res, grains_res)

    def test_smartos_zone_data(self):
        """
        Get basic information about a non-global zone
        """
        grains_exp_res = {
            "imageversion": "pkgbuild 18.1.0",
            "zoneid": "5",
            "zonename": "dda70f61-70fe-65e7-cf70-d878d69442d4",
        }

        cmd_mock = Mock(
            side_effect=[
                "5:dda70f61-70fe-65e7-cf70-d878d69442d4:running:/:dda70f61-70fe-65e7-cf70-d878d69442d4:native:excl:0",
            ]
        )
        fopen_mock = mock_open(
            read_data={
                "/etc/product": textwrap.dedent(
                    """\
                Name: Joyent Instance
                Image: pkgbuild 18.1.0
                Documentation: https://docs.joyent.com/images/smartos/pkgbuild
                """
                ),
            }
        )
        with patch.dict(smartos.__salt__, {"cmd.run": cmd_mock}), patch(
            "os.path.isfile", MagicMock(return_value=True)
        ), patch("salt.utils.files.fopen", fopen_mock):
            grains_res = smartos._smartos_zone_data()
            self.assertEqual(grains_exp_res, grains_res)

    def test_smartos_zone_pkgsrc_data_in_zone(self):
        """
        Get pkgsrc information from a zone
        """
        grains_exp_res = {
            "pkgsrcpath": (
                "https://pkgsrc.joyent.com/packages/SmartOS/2018Q1/x86_64/All"
            ),
            "pkgsrcversion": "2018Q1",
        }

        isfile_mock = Mock(side_effect=[True, False])
        fopen_mock = mock_open(
            read_data={
                "/opt/local/etc/pkg_install.conf": textwrap.dedent(
                    """\
                GPG_KEYRING_VERIFY=/opt/local/etc/gnupg/pkgsrc.gpg
                GPG_KEYRING_PKGVULN=/opt/local/share/gnupg/pkgsrc-security.gpg
                PKG_PATH=https://pkgsrc.joyent.com/packages/SmartOS/2018Q1/x86_64/All
                """
                ),
            }
        )

        with patch("os.path.isfile", isfile_mock), patch(
            "salt.utils.files.fopen", fopen_mock
        ):
            grains_res = smartos._smartos_zone_pkgsrc_data()
            self.assertEqual(grains_exp_res, grains_res)

    def test_smartos_zone_pkgsrc_data_in_globalzone(self):
        """
        Get pkgsrc information from the globalzone
        """
        grains_exp_res = {
            "pkgsrcpath": "https://pkgsrc.joyent.com/packages/SmartOS/trunk/tools/All",
            "pkgsrcversion": "trunk",
        }

        isfile_mock = Mock(side_effect=[False, True])
        fopen_mock = mock_open(
            read_data={
                "/opt/tools/etc/pkg_install.conf": textwrap.dedent(
                    """\
                GPG_KEYRING_PKGVULN=/opt/tools/share/gnupg/pkgsrc-security.gpg
                GPG_KEYRING_VERIFY=/opt/tools/etc/gnupg/pkgsrc.gpg
                PKG_PATH=https://pkgsrc.joyent.com/packages/SmartOS/trunk/tools/All
                VERIFIED_INSTALLATION=always
                """
                ),
            }
        )

        with patch("os.path.isfile", isfile_mock), patch(
            "salt.utils.files.fopen", fopen_mock
        ):
            grains_res = smartos._smartos_zone_pkgsrc_data()
            self.assertEqual(grains_exp_res, grains_res)

    def test_smartos_zone_pkgin_data_in_zone(self):
        """
        Get pkgin information from a zone
        """
        grains_exp_res = {
            "pkgin_repositories": [
                "https://pkgsrc.joyent.com/packages/SmartOS/2018Q1/x86_64/All",
                "http://pkg.blackdot.be/packages/2018Q1/x86_64/All",
            ],
        }

        isfile_mock = Mock(side_effect=[True, False])
        fopen_mock = mock_open(
            read_data={
                "/opt/local/etc/pkgin/repositories.conf": textwrap.dedent(
                    """\
                # $Id: repositories.conf,v 1.3 2012/06/13 13:50:17 imilh Exp $
                #
                # Pkgin repositories list
                #
                # Simply add repositories URIs one below the other
                #
                # WARNING: order matters, duplicates will not be added, if two
                # repositories hold the same package, it will be fetched from
                # the first one listed in this file.
                #
                # This file format supports the following macros:
                # $arch to define the machine hardware platform
                # $osrelease to define the release version for the operating system
                #
                # Remote ftp repository
                #
                # ftp://ftp.netbsd.org/pub/pkgsrc/packages/NetBSD/$arch/5.1/All
                #
                # Remote http repository
                #
                # http://mirror-master.dragonflybsd.org/packages/$arch/DragonFly-$osrelease/stable/All
                #
                # Local repository (must contain a pkg_summary.gz or bz2)
                #
                # file:///usr/pkgsrc/packages/All
                #
                https://pkgsrc.joyent.com/packages/SmartOS/2018Q1/x86_64/All
                http://pkg.blackdot.be/packages/2018Q1/x86_64/All
                """
                ),
            }
        )

        with patch("os.path.isfile", isfile_mock), patch(
            "salt.utils.files.fopen", fopen_mock
        ):
            grains_res = smartos._smartos_zone_pkgin_data()
            self.assertEqual(grains_exp_res, grains_res)

    def test_smartos_zone_pkgin_data_in_globalzone(self):
        """
        Get pkgin information from the globalzone
        """
        grains_exp_res = {
            "pkgin_repositories": [
                "https://pkgsrc.joyent.com/packages/SmartOS/trunk/tools/All",
            ],
        }

        isfile_mock = Mock(side_effect=[False, True])
        fopen_mock = mock_open(
            read_data={
                "/opt/tools/etc/pkgin/repositories.conf": textwrap.dedent(
                    """\
                #
                # Pkgin repositories list
                #
                # Simply add repositories URIs one below the other
                #
                # WARNING: order matters, duplicates will not be added, if two
                # repositories hold the same package, it will be fetched from
                # the first one listed in this file.
                #
                # This file format supports the following macros:
                # $arch to define the machine hardware platform
                # $osrelease to define the release version for the operating system
                #
                # Remote ftp repository
                #
                # ftp://ftp.netbsd.org/pub/pkgsrc/packages/NetBSD/$arch/5.1/All
                #
                # Remote http repository
                #
                # http://mirror-master.dragonflybsd.org/packages/$arch/DragonFly-$osrelease/stable/All
                #
                # Local repository (must contain a pkg_summary.gz or bz2)
                #
                # file:///usr/pkgsrc/packages/All
                #
                https://pkgsrc.joyent.com/packages/SmartOS/trunk/tools/All
                """
                ),
            }
        )

        with patch("os.path.isfile", isfile_mock), patch(
            "salt.utils.files.fopen", fopen_mock
        ):
            grains_res = smartos._smartos_zone_pkgin_data()
            self.assertEqual(grains_exp_res, grains_res)



import json
import pathlib

import pytest
import salt.states.ansiblegate as ansiblegate
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS


@pytest.fixture
def configure_loader_modules():
    return {ansiblegate: {}}


@pytest.fixture
def playbooks_examples_dir():
    return pathlib.Path(RUNTIME_VARS.TESTS_DIR) / "unit" / "files" / "playbooks"


def test_ansible_playbooks_states_success(playbooks_examples_dir):
    """
    Test ansible.playbooks states executions success.
    """

    success_output = json.loads(
        playbooks_examples_dir.joinpath("success_example.json").read_text()
    )

    with patch.dict(
        ansiblegate.__salt__,
        {"ansible.playbooks": MagicMock(return_value=success_output)},
    ), patch("salt.utils.path.which", return_value=True), patch.dict(
        ansiblegate.__opts__, {"test": False}
    ):
        ret = ansiblegate.playbooks("foobar")
        assert ret["result"] is True
        assert ret["comment"] == "Changes were made by playbook foobar"
        assert ret["changes"] == {
            "py2hosts": {
                "Ansible copy file to remote server": {"centos7-host1.tf.local": {}}
            }
        }


def test_ansible_playbooks_states_failed(playbooks_examples_dir):
    """
    Test ansible.playbooks failed states executions.
    :return:
    """
    failed_output = json.loads(
        playbooks_examples_dir.joinpath("failed_example.json").read_text()
    )
    with patch.dict(
        ansiblegate.__salt__,
        {"ansible.playbooks": MagicMock(return_value=failed_output)},
    ), patch("salt.utils.path.which", return_value=True), patch.dict(
        ansiblegate.__opts__, {"test": False}
    ):
        ret = ansiblegate.playbooks("foobar")
        assert ret["result"] is False
        assert ret["comment"] == "There were some issues running the playbook foobar"
        assert ret["changes"] == {
            "py2hosts": {
                "yum": {
                    "centos7-host1.tf.local": [
                        "No package matching 'rsyndc' found available, installed or"
                        " updated"
                    ]
                }
            }
        }



"""
Module for working with Windows PowerShell DSC (Desired State Configuration)

This module is Alpha

This module applies DSC Configurations in the form of PowerShell scripts or
MOF (Managed Object Format) schema files.

Use the ``psget`` module to manage PowerShell resources.

The idea is to leverage Salt to push DSC configuration scripts or MOF files to
the Minion.

:depends:
    - PowerShell 5.0
"""

import logging
import os

import salt.utils.json
import salt.utils.platform
import salt.utils.versions
from salt.exceptions import CommandExecutionError, SaltInvocationError

# Set up logging
log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "dsc"


def __virtual__():
    """
    Set the system module of the kernel is Windows
    """
    # Verify Windows
    if not salt.utils.platform.is_windows():
        log.debug("DSC: Only available on Windows systems")
        return False, "DSC: Only available on Windows systems"

    # Verify PowerShell
    powershell_info = __salt__["cmd.shell_info"]("powershell")
    if not powershell_info["installed"]:
        log.debug("DSC: Requires PowerShell")
        return False, "DSC: Requires PowerShell"

    # Verify PowerShell 5.0 or greater
    if salt.utils.versions.compare(powershell_info["version"], "<", "5.0"):
        log.debug("DSC: Requires PowerShell 5 or later")
        return False, "DSC: Requires PowerShell 5 or later"

    return __virtualname__


def _pshell(cmd, cwd=None, json_depth=2, ignore_retcode=False):
    """
    Execute the desired PowerShell command and ensure that it returns data
    in json format and load that into python. Either return a dict or raise a
    CommandExecutionError.
    """
    if "convertto-json" not in cmd.lower():
        cmd = "{} | ConvertTo-Json -Depth {}".format(cmd, json_depth)
    log.debug("DSC: %s", cmd)
    results = __salt__["cmd.run_all"](
        cmd,
        shell="powershell",
        cwd=cwd,
        python_shell=True,
        ignore_retcode=ignore_retcode,
    )

    if "pid" in results:
        del results["pid"]

    if "retcode" not in results or results["retcode"] != 0:
        # run_all logs an error to log.error, fail hard back to the user
        raise CommandExecutionError(
            "Issue executing PowerShell {}".format(cmd), info=results
        )

    # Sometimes Powershell returns an empty string, which isn't valid JSON
    if results["stdout"] == "":
        results["stdout"] = "{}"

    try:
        ret = salt.utils.json.loads(results["stdout"], strict=False)
    except ValueError:
        raise CommandExecutionError("No JSON results from PowerShell", info=results)

    log.info('DSC: Returning "%s"', ret)
    return ret


def run_config(
    path,
    source=None,
    config_name=None,
    config_data=None,
    config_data_source=None,
    script_parameters=None,
    salt_env="base",
):
    r"""
    Compile a DSC Configuration in the form of a PowerShell script (.ps1) and
    apply it. The PowerShell script can be cached from the master using the
    ``source`` option. If there is more than one config within the PowerShell
    script, the desired configuration can be applied by passing the name in the
    ``config`` option.

    This command would be the equivalent of running ``dsc.compile_config``
    followed by ``dsc.apply_config``.

    Args:

        path (str): The local path to the PowerShell script that contains the
            DSC Configuration. Required.

        source (str): The path to the script on ``file_roots`` to cache at the
            location specified by ``path``. The source file will be cached
            locally and then executed. If source is not passed, the config
            script located at ``path`` will be compiled. Optional.

        config_name (str): The name of the Configuration within the script to
            apply. If the script contains multiple configurations within the
            file a ``config_name`` must be specified. If the ``config_name`` is
            not specified, the name of the file will be used as the
            ``config_name`` to run. Optional.

        config_data (str): Configuration data in the form of a hash table that
            will be passed to the ``ConfigurationData`` parameter when the
            ``config_name`` is compiled. This can be the path to a ``.psd1``
            file containing the proper hash table or the PowerShell code to
            create the hash table.

            .. versionadded:: 2017.7.0

        config_data_source (str): The path to the ``.psd1`` file on
            ``file_roots`` to cache at the location specified by
            ``config_data``. If this is specified, ``config_data`` must be a
            local path instead of a hash table.

            .. versionadded:: 2017.7.0

        script_parameters (str): Any additional parameters expected by the
            configuration script. These must be defined in the script itself.

            .. versionadded:: 2017.7.0

        salt_env (str): The salt environment to use when copying the source.
            Default is 'base'

    Returns:
        bool: True if successfully compiled and applied, otherwise False

    CLI Example:

    To compile a config from a script that already exists on the system:

    .. code-block:: bash

        salt '*' dsc.run_config C:\\DSC\\WebsiteConfig.ps1

    To cache a config script to the system from the master and compile it:

    .. code-block:: bash

        salt '*' dsc.run_config C:\\DSC\\WebsiteConfig.ps1 salt://dsc/configs/WebsiteConfig.ps1
    """
    ret = compile_config(
        path=path,
        source=source,
        config_name=config_name,
        config_data=config_data,
        config_data_source=config_data_source,
        script_parameters=script_parameters,
        salt_env=salt_env,
    )

    if ret.get("Exists"):
        config_path = os.path.dirname(ret["FullName"])
        return apply_config(config_path)
    else:
        return False


def compile_config(
    path,
    source=None,
    config_name=None,
    config_data=None,
    config_data_source=None,
    script_parameters=None,
    salt_env="base",
):
    r"""
    Compile a config from a PowerShell script (``.ps1``)

    Args:

        path (str): Path (local) to the script that will create the ``.mof``
            configuration file. If no source is passed, the file must exist
            locally. Required.

        source (str): Path to the script on ``file_roots`` to cache at the
            location specified by ``path``. The source file will be cached
            locally and then executed. If source is not passed, the config
            script located at ``path`` will be compiled. Optional.

        config_name (str): The name of the Configuration within the script to
            apply. If the script contains multiple configurations within the
            file a ``config_name`` must be specified. If the ``config_name`` is
            not specified, the name of the file will be used as the
            ``config_name`` to run. Optional.

        config_data (str): Configuration data in the form of a hash table that
            will be passed to the ``ConfigurationData`` parameter when the
            ``config_name`` is compiled. This can be the path to a ``.psd1``
            file containing the proper hash table or the PowerShell code to
            create the hash table.

            .. versionadded:: 2017.7.0

        config_data_source (str): The path to the ``.psd1`` file on
            ``file_roots`` to cache at the location specified by
            ``config_data``. If this is specified, ``config_data`` must be a
            local path instead of a hash table.

            .. versionadded:: 2017.7.0

        script_parameters (str): Any additional parameters expected by the
            configuration script. These must be defined in the script itself.

            .. versionadded:: 2017.7.0

        salt_env (str): The salt environment to use when copying the source.
            Default is 'base'

    Returns:
        dict: A dictionary containing the results of the compilation

    CLI Example:

    To compile a config from a script that already exists on the system:

    .. code-block:: bash

        salt '*' dsc.compile_config C:\\DSC\\WebsiteConfig.ps1

    To cache a config script to the system from the master and compile it:

    .. code-block:: bash

        salt '*' dsc.compile_config C:\\DSC\\WebsiteConfig.ps1 salt://dsc/configs/WebsiteConfig.ps1
    """
    if source:
        log.info("DSC: Caching %s", source)
        cached_files = __salt__["cp.get_file"](
            path=source, dest=path, saltenv=salt_env, makedirs=True
        )
        if not cached_files:
            error = "Failed to cache {}".format(source)
            log.error("DSC: %s", error)
            raise CommandExecutionError(error)

    if config_data_source:
        log.info("DSC: Caching %s", config_data_source)
        cached_files = __salt__["cp.get_file"](
            path=config_data_source, dest=config_data, saltenv=salt_env, makedirs=True
        )
        if not cached_files:
            error = "Failed to cache {}".format(config_data_source)
            log.error("DSC: %s", error)
            raise CommandExecutionError(error)

    # Make sure the path exists
    if not os.path.exists(path):
        error = '"{}" not found'.format(path)
        log.error("DSC: %s", error)
        raise CommandExecutionError(error)

    if config_name is None:
        # If the name of the config isn't passed, make it the name of the .ps1
        config_name = os.path.splitext(os.path.basename(path))[0]

    cwd = os.path.dirname(path)

    # Run the script and see if the compile command is in the script
    cmd = [path]
    # Add any script parameters
    if script_parameters:
        cmd.append(script_parameters)
    # Select fields to return
    cmd.append(
        "| Select-Object -Property FullName, Extension, Exists, "
        '@{Name="LastWriteTime";Expression={Get-Date ($_.LastWriteTime) '
        "-Format g}}"
    )

    cmd = " ".join(cmd)

    ret = _pshell(cmd, cwd)

    if ret:
        # Script compiled, return results
        if ret.get("Exists"):
            log.info("DSC: Compile Config: %s", ret)
            return ret

    # If you get to this point, the script did not contain a compile command
    # dot source the script to compile the state and generate the mof file
    cmd = [".", path]
    if script_parameters:
        cmd.append(script_parameters)
    cmd.extend([";", config_name])
    if config_data:
        cmd.extend(["-ConfigurationData", config_data])
    cmd.append(
        "| Select-Object -Property FullName, Extension, Exists, "
        '@{Name="LastWriteTime";Expression={Get-Date ($_.LastWriteTime) '
        "-Format g}}"
    )

    cmd = " ".join(cmd)

    ret = _pshell(cmd, cwd)

    if ret:
        # Script compiled, return results
        if ret.get("Exists"):
            log.info("DSC: Compile Config: %s", ret)
            return ret

    error = "Failed to compile config: {}".format(path)
    error += "\nReturned: {}".format(ret)
    log.error("DSC: %s", error)
    raise CommandExecutionError(error)


def apply_config(path, source=None, salt_env="base"):
    r"""
    Run an compiled DSC configuration (a folder containing a .mof file). The
    folder can be cached from the salt master using the ``source`` option.

    Args:

        path (str): Local path to the directory that contains the .mof
            configuration file to apply. Required.

        source (str): Path to the directory that contains the .mof file on the
            ``file_roots``. The source directory will be copied to the path
            directory and then executed. If the path and source directories
            differ, the source directory will be applied. If source is not
            passed, the config located at ``path`` will be applied. Optional.

        salt_env (str): The salt environment to use when copying your source.
            Default is 'base'

    Returns:
        bool: True if successful, otherwise False

    CLI Example:

    To apply a config that already exists on the system

    .. code-block:: bash

        salt '*' dsc.apply_config C:\\DSC\\WebSiteConfiguration

    To cache a configuration from the master and apply it:

    .. code-block:: bash

        salt '*' dsc.apply_config C:\\DSC\\WebSiteConfiguration salt://dsc/configs/WebSiteConfiguration

    """
    # If you're getting an error along the lines of "The client cannot connect
    # to the destination specified in the request.", try the following:
    # Enable-PSRemoting -SkipNetworkProfileCheck
    config = path
    if source:
        # Make sure the folder names match
        path_name = os.path.basename(os.path.normpath(path))
        source_name = os.path.basename(os.path.normpath(source))
        if path_name.lower() != source_name.lower():
            # Append the Source name to the Path
            path = "{}\\{}".format(path, source_name)
            log.debug("DSC: %s appended to the path.", source_name)

        # Destination path minus the basename
        dest_path = os.path.dirname(os.path.normpath(path))
        log.info("DSC: Caching %s", source)
        cached_files = __salt__["cp.get_dir"](source, dest_path, salt_env)
        if not cached_files:
            error = "Failed to copy {}".format(source)
            log.error("DSC: %s", error)
            raise CommandExecutionError(error)
        else:
            config = os.path.dirname(cached_files[0])

    # Make sure the path exists
    if not os.path.exists(config):
        error = "{} not found".format(config)
        log.error("DSC: %s", error)
        raise CommandExecutionError(error)

    # Run the DSC Configuration
    # Putting quotes around the parameter protects against command injection
    cmd = 'Start-DscConfiguration -Path "{}" -Wait -Force'.format(config)
    _pshell(cmd)

    cmd = "$status = Get-DscConfigurationStatus; $status.Status"
    ret = _pshell(cmd)
    log.info("DSC: Apply Config: %s", ret)

    return ret == "Success" or ret == {}


def get_config():
    """
    Get the current DSC Configuration

    Returns:
        dict: A dictionary representing the DSC Configuration on the machine

    Raises:
        CommandExecutionError: On failure

    CLI Example:

    .. code-block:: bash

        salt '*' dsc.get_config
    """
    cmd = "Get-DscConfiguration | Select-Object * -ExcludeProperty Cim*"

    try:
        raw_config = _pshell(cmd, ignore_retcode=True)
    except CommandExecutionError as exc:
        if "Current configuration does not exist" in exc.info["stderr"]:
            raise CommandExecutionError("Not Configured")
        raise

    config = dict()
    if raw_config:
        # Get DSC Configuration Name
        if "ConfigurationName" in raw_config[0]:
            config[raw_config[0]["ConfigurationName"]] = {}
        # Add all DSC Configurations by ResourceId
        for item in raw_config:
            config[item["ConfigurationName"]][item["ResourceId"]] = {}
            for key in item:
                if key not in ["ConfigurationName", "ResourceId"]:
                    config[item["ConfigurationName"]][item["ResourceId"]][key] = item[
                        key
                    ]

    return config


def remove_config(reset=False):
    """
    Remove the current DSC Configuration. Removes current, pending, and previous
    dsc configurations.

    .. versionadded:: 2017.7.5

    Args:
        reset (bool):
            Attempts to reset the DSC configuration by removing the following
            from ``C:\\Windows\\System32\\Configuration``:

            - File: DSCStatusHistory.mof
            - File: DSCEngineCache.mof
            - Dir: ConfigurationStatus

            Default is False

            .. warning::
                ``remove_config`` may fail to reset the DSC environment if any
                of the files in the ``ConfigurationStatus`` directory. If you
                wait a few minutes and run again, it may complete successfully.

    Returns:
        bool: True if successful

    Raises:
        CommandExecutionError: On failure

    CLI Example:

    .. code-block:: bash

        salt '*' dsc.remove_config True
    """
    # Stopping a running config (not likely to occur)
    cmd = "Stop-DscConfiguration"
    log.info("DSC: Stopping Running Configuration")
    try:
        _pshell(cmd)
    except CommandExecutionError as exc:
        if exc.info["retcode"] != 0:
            raise CommandExecutionError(
                "Failed to Stop DSC Configuration", info=exc.info
            )
        log.info("DSC: %s", exc.info["stdout"])

    # Remove configuration files
    cmd = "Remove-DscConfigurationDocument -Stage Current, Pending, Previous -Force"
    log.info("DSC: Removing Configuration")
    try:
        _pshell(cmd)
    except CommandExecutionError as exc:
        if exc.info["retcode"] != 0:
            raise CommandExecutionError(
                "Failed to remove DSC Configuration", info=exc.info
            )
        log.info("DSC: %s", exc.info["stdout"])

    if not reset:
        return True

    def _remove_fs_obj(path):
        if os.path.exists(path):
            log.info("DSC: Removing %s", path)
            if not __salt__["file.remove"](path):
                error = "Failed to remove {}".format(path)
                log.error("DSC: %s", error)
                raise CommandExecutionError(error)

    dsc_config_dir = "{}\\System32\\Configuration".format(
        os.getenv("SystemRoot", "C:\\Windows")
    )

    # Remove History
    _remove_fs_obj("{}\\DSCStatusHistory.mof".format(dsc_config_dir))

    # Remove Engine Cache
    _remove_fs_obj("{}\\DSCEngineCache.mof".format(dsc_config_dir))

    # Remove Status Directory
    _remove_fs_obj("{}\\ConfigurationStatus".format(dsc_config_dir))

    return True


def restore_config():
    """
    Reapplies the previous configuration.

    .. versionadded:: 2017.7.5

    .. note::
        The current configuration will be come the previous configuration. If
        run a second time back-to-back it is like toggling between two configs.

    Returns:
        bool: True if successfully restored

    Raises:
        CommandExecutionError: On failure

    CLI Example:

    .. code-block:: bash

        salt '*' dsc.restore_config
    """
    cmd = "Restore-DscConfiguration"
    try:
        _pshell(cmd, ignore_retcode=True)
    except CommandExecutionError as exc:
        if "A previous configuration does not exist" in exc.info["stderr"]:
            raise CommandExecutionError("Previous Configuration Not Found")
        raise
    return True


def test_config():
    """
    Tests the current applied DSC Configuration

    Returns:
        bool: True if successfully applied, otherwise False

    CLI Example:

    .. code-block:: bash

        salt '*' dsc.test_config
    """
    cmd = "Test-DscConfiguration"
    try:
        _pshell(cmd, ignore_retcode=True)
    except CommandExecutionError as exc:
        if "Current configuration does not exist" in exc.info["stderr"]:
            raise CommandExecutionError("Not Configured")
        raise


def get_config_status():
    """
    Get the status of the current DSC Configuration

    Returns:
        dict: A dictionary representing the status of the current DSC
            Configuration on the machine

    CLI Example:

    .. code-block:: bash

        salt '*' dsc.get_config_status
    """
    cmd = (
        "Get-DscConfigurationStatus | "
        "Select-Object -Property HostName, Status, MetaData, "
        '@{Name="StartDate";Expression={Get-Date ($_.StartDate) -Format g}}, '
        "Type, Mode, RebootRequested, NumberofResources"
    )
    try:
        return _pshell(cmd, ignore_retcode=True)
    except CommandExecutionError as exc:
        if "No status information available" in exc.info["stderr"]:
            raise CommandExecutionError("Not Configured")
        raise


def get_lcm_config():
    """
    Get the current Local Configuration Manager settings

    Returns:
        dict: A dictionary representing the Local Configuration Manager settings
            on the machine

    CLI Example:

    .. code-block:: bash

        salt '*' dsc.get_lcm_config
    """
    cmd = (
        "Get-DscLocalConfigurationManager | "
        "Select-Object -Property ConfigurationModeFrequencyMins, LCMState, "
        "RebootNodeIfNeeded, ConfigurationMode, ActionAfterReboot, "
        "RefreshMode, CertificateID, ConfigurationID, RefreshFrequencyMins, "
        "AllowModuleOverwrite, DebugMode, StatusRetentionTimeInDays "
    )
    return _pshell(cmd)


def set_lcm_config(
    config_mode=None,
    config_mode_freq=None,
    refresh_freq=None,
    reboot_if_needed=None,
    action_after_reboot=None,
    refresh_mode=None,
    certificate_id=None,
    configuration_id=None,
    allow_module_overwrite=None,
    debug_mode=False,
    status_retention_days=None,
):
    """
    For detailed descriptions of the parameters see:
    https://msdn.microsoft.com/en-us/PowerShell/DSC/metaConfig

    config_mode (str): How the LCM applies the configuration. Valid values
        are:

        - ApplyOnly
        - ApplyAndMonitor
        - ApplyAndAutoCorrect

    config_mode_freq (int): How often, in minutes, the current configuration
        is checked and applied. Ignored if config_mode is set to ApplyOnly.
        Default is 15.

    refresh_mode (str): How the LCM gets configurations. Valid values are:

        - Disabled
        - Push
        - Pull

    refresh_freq (int): How often, in minutes, the LCM checks for updated
        configurations. (pull mode only) Default is 30.

    reboot_if_needed (bool): Reboot the machine if needed after a
        configuration is applied. Default is False.

    action_after_reboot (str): Action to take after reboot. Valid values
        are:

        - ContinueConfiguration
        - StopConfiguration

    certificate_id (guid): A GUID that specifies a certificate used to
        access the configuration: (pull mode)

    configuration_id (guid): A GUID that identifies the config file to get
        from a pull server. (pull mode)

    allow_module_overwrite (bool): New configs are allowed to overwrite old
        ones on the target node.

    debug_mode (str): Sets the debug level. Valid values are:

        - None
        - ForceModuleImport
        - All

    status_retention_days (int): Number of days to keep status of the
        current config.

    .. note::
        Either ``config_mode_freq`` or ``refresh_freq`` needs to be a
        multiple of the other. See documentation on MSDN for more details.

    Returns:
        bool: True if successful, otherwise False

    CLI Example:

    .. code-block:: bash

        salt '*' dsc.set_lcm_config ApplyOnly
    """
    temp_dir = os.getenv("TEMP", "{}\\temp".format(os.getenv("WINDIR")))
    cmd = "Configuration SaltConfig {"
    cmd += "    Node localhost {"
    cmd += "        LocalConfigurationManager {"
    if config_mode:
        if config_mode not in ("ApplyOnly", "ApplyAndMonitor", "ApplyAndAutoCorrect"):
            error = (
                "config_mode must be one of ApplyOnly, ApplyAndMonitor, "
                "or ApplyAndAutoCorrect. Passed {}".format(config_mode)
            )
            raise SaltInvocationError(error)
        cmd += '            ConfigurationMode = "{}";'.format(config_mode)
    if config_mode_freq:
        if not isinstance(config_mode_freq, int):
            error = "config_mode_freq must be an integer. Passed {}".format(
                config_mode_freq
            )
            raise SaltInvocationError(error)
        cmd += "            ConfigurationModeFrequencyMins = {};".format(
            config_mode_freq
        )
    if refresh_mode:
        if refresh_mode not in ("Disabled", "Push", "Pull"):
            raise SaltInvocationError(
                "refresh_mode must be one of Disabled, Push, or Pull"
            )
        cmd += '            RefreshMode = "{}";'.format(refresh_mode)
    if refresh_freq:
        if not isinstance(refresh_freq, int):
            raise SaltInvocationError("refresh_freq must be an integer")
        cmd += "            RefreshFrequencyMins = {};".format(refresh_freq)
    if reboot_if_needed is not None:
        if not isinstance(reboot_if_needed, bool):
            raise SaltInvocationError("reboot_if_needed must be a boolean value")
        if reboot_if_needed:
            reboot_if_needed = "$true"
        else:
            reboot_if_needed = "$false"
        cmd += "            RebootNodeIfNeeded = {};".format(reboot_if_needed)
    if action_after_reboot:
        if action_after_reboot not in ("ContinueConfiguration", "StopConfiguration"):
            raise SaltInvocationError(
                "action_after_reboot must be one of "
                "ContinueConfiguration or StopConfiguration"
            )
        cmd += '            ActionAfterReboot = "{}"'.format(action_after_reboot)
    if certificate_id is not None:
        if certificate_id == "":
            certificate_id = None
        cmd += '            CertificateID = "{}";'.format(certificate_id)
    if configuration_id is not None:
        if configuration_id == "":
            configuration_id = None
        cmd += '            ConfigurationID = "{}";'.format(configuration_id)
    if allow_module_overwrite is not None:
        if not isinstance(allow_module_overwrite, bool):
            raise SaltInvocationError("allow_module_overwrite must be a boolean value")
        if allow_module_overwrite:
            allow_module_overwrite = "$true"
        else:
            allow_module_overwrite = "$false"
        cmd += "            AllowModuleOverwrite = {};".format(allow_module_overwrite)
    if debug_mode is not False:
        if debug_mode is None:
            debug_mode = "None"
        if debug_mode not in ("None", "ForceModuleImport", "All"):
            raise SaltInvocationError(
                "debug_mode must be one of None, ForceModuleImport, "
                "ResourceScriptBreakAll, or All"
            )
        cmd += '            DebugMode = "{}";'.format(debug_mode)
    if status_retention_days:
        if not isinstance(status_retention_days, int):
            raise SaltInvocationError("status_retention_days must be an integer")
        cmd += "            StatusRetentionTimeInDays = {};".format(
            status_retention_days
        )
    cmd += "        }}};"
    cmd += r'SaltConfig -OutputPath "{}\SaltConfig"'.format(temp_dir)

    # Execute Config to create the .mof
    _pshell(cmd)

    # Apply the config
    cmd = r'Set-DscLocalConfigurationManager -Path "{}\SaltConfig"' r"".format(temp_dir)
    ret = __salt__["cmd.run_all"](cmd, shell="powershell", python_shell=True)
    __salt__["file.remove"](r"{}\SaltConfig".format(temp_dir))
    if not ret["retcode"]:
        log.info("DSC: LCM config applied successfully")
        return True
    else:
        log.error("DSC: Failed to apply LCM config. Error %s", ret)
        return False
